from .base import BaseAdapter
from .litellm_adapter import LiteLLMAdapter

__all__ = ["BaseAdapter", "get_adapter", "LiteLLMAdapter"]


def get_adapter(model, api_key=None, api_base=None):
    return LiteLLMAdapter(model, api_key, api_base)



import ast
import fnmatch
from collections import defaultdict
from pathlib import Path
from typing import Any
from collections.abc import Iterator

FRAMEWORK_DECORATORS = [
    "@*.route",
    "@*.get",
    "@*.post",
    "@*.put",
    "@*.delete",
    "@*.patch",
    "@*.before_request",
    "@*.after_request",
    "@*.errorhandler",
    "@*.teardown_*",
    "@*.head",
    "@*.options",
    "@*.trace",
    "@*.websocket",
    "@*.middleware",
    "@*.on_event",
    "@*.exception_handler",
    "@*_required",
    "@login_required",
    "@permission_required",
    "django.views.decorators.*",
    "@*.simple_tag",
    "@*.inclusion_tag",
    "@*.filter",
    "@*.tag",
    "@*.register",
    "@validator",
    "@field_validator",
    "@model_validator",
    "@root_validator",
    "@field_serializer",
    "@model_serializer",
    "@computed_field",
    "@*.command",
    "@*.default",
    "@*.callback",
    "@*.group",
    "@*.subcommand",
    "@*.main",
    "@shared_task",
    "@*.shared_task",
    "@*.task",
    "@*.signal",
    "@*.lifespan",
    "@pre_load",
    "@post_load",
    "@pre_dump",
    "@post_dump",
    "@validates",
    "@validates_schema",
    "@*.pre_load",
    "@*.post_load",
    "@*.pre_dump",
    "@*.post_dump",
    "@*.validates",
    "@*.validates_schema",
    "@*.listens_for",
    "@listens_for",
]

FRAMEWORK_FUNCTIONS = [
    "get",
    "post",
    "put",
    "patch",
    "delete",
    "head",
    "options",
    "trace",
    "*_queryset",
    "get_queryset",
    "get_object",
    "get_context_data",
    "*_form",
    "form_valid",
    "form_invalid",
    "get_form_*",
]

ENTRY_POINT_DECORATORS = {
    "app.route",
    "app.get",
    "app.post",
    "app.put",
    "app.delete",
    "router.get",
    "router.post",
    "router.put",
    "router.delete",
    "blueprint.route",
    "blueprint.get",
    "blueprint.post",
    "celery.task",
    "shared_task",
    "task",
    "job",
    "click.command",
    "command",
    "pytest.fixture",
    "fixture",
    "receiver",
    "admin.register",
    "on_event",
    "subscriber",
    "listener",
    "handler",
    "app.before_first_request",
    "app.cli.command",
}

FRAMEWORK_IMPORTS = {
    "flask",
    "fastapi",
    "django",
    "rest_framework",
    "pydantic",
    "celery",
    "starlette",
    "uvicorn",
    "marshmallow",
    "tornado",
    "sanic",
    "aiohttp",
    "falcon",
    "bottle",
    "typer",
    "click",
    "sqlalchemy",
}

ROUTE_METHODS = {
    "route",
    "get",
    "post",
    "put",
    "delete",
    "patch",
    "head",
    "options",
    "trace",
    "websocket",
}


class FrameworkAwareVisitor:
    def __init__(self, filename: str | Path | None = None) -> None:
        self.is_framework_file = False
        self.detected_frameworks = set()
        self.framework_decorated_lines = set()
        self.func_defs = {}
        self.class_defs = {}
        self.class_method_lines = {}
        self.pydantic_models = set()
        self._mark_functions = set()
        self._mark_classes = set()
        self.declarative_classes = set()
        self._mark_cbv_http_methods = set()
        self._type_refs_in_routes = set()
        self.objects_with_routes = defaultdict(list)
        self.objects_passed_as_args = set()
        self.objects_created_by_call = set()

        if filename:
            self._check_framework_imports_in_file(filename)

    def visit(self, node: ast.AST) -> Any:
        method = "visit_" + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)

    def generic_visit(self, node: ast.AST) -> None:
        for field, value in ast.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        self.visit(item)
            elif isinstance(value, ast.AST):
                self.visit(value)

    def visit_Import(self, node: ast.Import) -> None:
        for alias in node.names:
            name = alias.name.lower()

            for fw in FRAMEWORK_IMPORTS:
                if fw in name:
                    self.is_framework_file = True
                    framework_name = name.split(".")[0]
                    self.detected_frameworks.add(framework_name)
                    break

        self.generic_visit(node)

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        if node.module:
            module_name = node.module.split(".")[0].lower()
            if module_name in FRAMEWORK_IMPORTS:
                self.is_framework_file = True
                self.detected_frameworks.add(module_name)
        self.generic_visit(node)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self.func_defs.setdefault(node.name, node.lineno)
        is_route = False

        for deco in node.decorator_list:
            d = self._normalize_decorator(deco)

            router_name = self._get_router_from_decorator(deco)
            if router_name:
                self.objects_with_routes[router_name].append(node.lineno)
                self.is_framework_file = True
                is_route = True

            if self._matches_framework_pattern(d, FRAMEWORK_DECORATORS):
                self.is_framework_file = True
                self.framework_decorated_lines.add(node.lineno)
                is_route = True

            if self._decorator_base_name_is(deco, "receiver"):
                self.framework_decorated_lines.add(node.lineno)
                self.is_framework_file = True
                is_route = True

        defaults_to_scan = []
        if node.args.defaults:
            defaults_to_scan.extend(node.args.defaults)
        if node.args.kw_defaults:
            defaults_to_scan.extend(node.args.kw_defaults)

        for default in defaults_to_scan:
            self._scan_for_depends(default)

        if is_route:
            self._collect_annotation_type_refs(node)
        self.generic_visit(node)

    visit_AsyncFunctionDef = visit_FunctionDef

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        self.class_defs[node.name] = node
        for item in node.body:
            if isinstance(item, (ast.FunctionDef, ast.AsyncFunctionDef)):
                self.class_method_lines[(node.name, item.name)] = item.lineno
        bases = self._base_names(node)

        is_view_like = False
        for base in bases:
            for token in ("view", "viewset", "apiview", "handler"):
                if token in base:
                    is_view_like = True
                    break
            if is_view_like:
                break

        is_pydantic = False
        for base in bases:
            if "basemodel" in base:
                is_pydantic = True
                break

        if is_view_like:
            self.is_framework_file = True
            self._mark_cbv_http_methods.add(node.name)

        if is_pydantic:
            self.pydantic_models.add(node.name)
            self.declarative_classes.add(node.name)
            self.is_framework_file = True

        else:
            for base in bases:
                tail = base.split(".")[-1]
                if tail in ("schema", "model"):
                    self.declarative_classes.add(node.name)
                    break

        self.generic_visit(node)

    def visit_Assign(self, node: ast.Assign) -> None:
        if isinstance(node.value, ast.Call):
            for target in node.targets:
                if isinstance(target, ast.Name):
                    self.objects_created_by_call.add(target.id)

        targets = []
        for t in node.targets:
            if isinstance(t, ast.Name):
                targets.append(t.id)

        if "urlpatterns" in targets:
            self.is_framework_file = True
            for elt in self._iter_list_elts(node.value):
                if isinstance(elt, ast.Call) and self._call_name_endswith(
                    elt, {"path", "re_path"}
                ):
                    view_expr = self._get_posarg(elt, 1)
                    self._mark_view_from_url_pattern(view_expr)
        self.generic_visit(node)

    def visit_Call(self, node: ast.Call) -> None:
        if isinstance(node.func, ast.Attribute) and node.func.attr == "register":
            if len(node.args) >= 2:
                vs = node.args[1]
                cls_name = self._simple_name(vs)
                if cls_name:
                    self._mark_classes.add(cls_name)
                    self._mark_cbv_http_methods.add(cls_name)
                    self.is_framework_file = True
        if (
            isinstance(node.func, ast.Attribute)
            and node.func.attr == "connect"
            and node.args
        ):
            func_name = self._simple_name(node.args[0])
            if func_name:
                self._mark_functions.add(func_name)
                self.is_framework_file = True

        for arg in node.args:
            if isinstance(arg, ast.Name):
                self.objects_passed_as_args.add(arg.id)
        for kw in node.keywords:
            if isinstance(kw.value, ast.Name):
                self.objects_passed_as_args.add(kw.value.id)

        self.generic_visit(node)

    def finalize(self) -> None:
        for fname in self._mark_functions:
            if fname in self.func_defs:
                self.framework_decorated_lines.add(self.func_defs[fname])
        for cname in self._mark_classes:
            cls_node = self.class_defs.get(cname)
            if cls_node is not None:
                self.framework_decorated_lines.add(cls_node.lineno)
        for cname in self._mark_cbv_http_methods:
            for meth in (
                "get",
                "post",
                "put",
                "patch",
                "delete",
                "head",
                "options",
                "trace",
                "list",
                "create",
                "retrieve",
                "update",
                "partial_update",
                "destroy",
            ):
                lino = self.class_method_lines.get((cname, meth))
                if lino:
                    self.framework_decorated_lines.add(lino)

        for obj_name, route_lines in self.objects_with_routes.items():
            for line in route_lines:
                self.framework_decorated_lines.add(line)

        typed_models = set()
        for t in self._type_refs_in_routes:
            if t in self.pydantic_models:
                typed_models.add(t)

        self._mark_classes.update(typed_models)
        for cname in typed_models:
            cls_node = self.class_defs.get(cname)
            if cls_node is not None:
                self.framework_decorated_lines.add(cls_node.lineno)

    def _check_framework_imports_in_file(self, filename: str | Path) -> None:
        try:
            content = Path(filename).read_text(encoding="utf-8")

            for framework in FRAMEWORK_IMPORTS:
                import_statement = f"import {framework}"
                from_statement = f"from {framework}"

                has_import = import_statement in content
                has_from_import = from_statement in content

                if has_import or has_from_import:
                    self.is_framework_file = True
                    self.detected_frameworks.add(framework)
                    break

        except Exception:
            pass

    def _normalize_decorator(self, dec: ast.AST) -> str:
        if isinstance(dec, ast.Call):
            return self._normalize_decorator(dec.func)
        if isinstance(dec, ast.Name):
            return f"@{dec.id}"
        if isinstance(dec, ast.Attribute):
            return f"@{self._attr_to_str(dec)}"
        return "@unknown"

    def _matches_framework_pattern(self, text: str, patterns: list[str]) -> bool:
        text_clean = text.lstrip("@")

        for pattern in patterns:
            pattern_clean = pattern.lstrip("@")
            if fnmatch.fnmatch(text_clean, pattern_clean):
                return True

        return False

    def _decorator_base_name_is(self, dec: ast.AST, name: str) -> bool:
        if isinstance(dec, ast.Call):
            dec = dec.func
        if isinstance(dec, ast.Name):
            return dec.id == name
        if isinstance(dec, ast.Attribute):
            return dec.attr == name or self._attr_to_str(dec).endswith("." + name)
        return False

    def _attr_to_str(self, node: ast.Attribute) -> str:
        parts = []
        cur = node
        while isinstance(cur, ast.Attribute):
            parts.append(cur.attr)
            cur = cur.value
        if isinstance(cur, ast.Name):
            parts.append(cur.id)

        parts.reverse()
        return ".".join(parts)

    def _base_names(self, node: ast.ClassDef) -> list[str]:
        out = []
        for b in node.bases:
            if isinstance(b, ast.Name):
                out.append(b.id.lower())
            elif isinstance(b, ast.Attribute):
                out.append(self._attr_to_str(b).lower())
        return out

    def _iter_list_elts(self, node: ast.AST) -> Iterator[ast.expr]:
        if isinstance(node, (ast.List, ast.Tuple, ast.Set)):
            for elt in node.elts:
                yield elt

    def _call_name_endswith(self, call: ast.Call, names: set[str]) -> bool:
        if isinstance(call.func, ast.Name):
            return call.func.id in names
        if isinstance(call.func, ast.Attribute):
            return call.func.attr in names
        return False

    def _get_posarg(self, call: ast.Call, idx: int) -> ast.expr | None:
        return call.args[idx] if len(call.args) > idx else None

    def _simple_name(self, node: ast.AST) -> str | None:
        if isinstance(node, ast.Name):
            return node.id
        if isinstance(node, ast.Attribute):
            return node.attr
        return None

    def _mark_view_from_url_pattern(self, view_expr: ast.AST | None) -> None:
        if view_expr is None:
            return
        if (
            isinstance(view_expr, ast.Call)
            and isinstance(view_expr.func, ast.Attribute)
            and view_expr.func.attr == "as_view"
        ):
            cls_name = self._simple_name(view_expr.func.value)
            if cls_name:
                self._mark_classes.add(cls_name)
                self._mark_cbv_http_methods.add(cls_name)
        else:
            fname = self._simple_name(view_expr)
            if fname:
                self._mark_functions.add(fname)

    def _scan_for_depends(self, node) -> None:
        if not isinstance(node, ast.Call):
            return
        is_depends = False
        if isinstance(node.func, ast.Name) and node.func.id == "Depends":
            is_depends = True
        elif isinstance(node.func, ast.Attribute) and node.func.attr == "Depends":
            is_depends = True
        if not is_depends:
            return
        if node.args:
            dep = node.args[0]
            dep_name = self._simple_name(dep)
            if dep_name:
                self._mark_functions.add(dep_name)
                self.is_framework_file = True

    def _collect_annotation_type_refs(self, fn: ast.FunctionDef) -> None:
        def collect(t):
            if t is None:
                return

            if isinstance(t, ast.Name):
                self._type_refs_in_routes.add(t.id)
                return

            if isinstance(t, ast.Attribute):
                self._type_refs_in_routes.add(t.attr)
                return

            if isinstance(t, ast.Subscript):
                collect(t.value)
                slice_node = t.slice
                if isinstance(slice_node, ast.Tuple):
                    for element in slice_node.elts:
                        collect(element)
                else:
                    collect(slice_node)
                return

            if isinstance(t, ast.Tuple):
                for element in t.elts:
                    collect(element)

        all_args = []
        all_args.extend(fn.args.args)
        all_args.extend(fn.args.posonlyargs)
        all_args.extend(fn.args.kwonlyargs)

        for arg in all_args:
            collect(arg.annotation)

        if fn.returns:
            collect(fn.returns)

    def _get_router_from_decorator(self, deco: ast.AST) -> str | None:
        if isinstance(deco, ast.Call):
            deco = deco.func

        if isinstance(deco, ast.Attribute):
            if deco.attr in ROUTE_METHODS:
                if isinstance(deco.value, ast.Name):
                    return deco.value.id
        return None


def detect_framework_usage(
    definition: Any, visitor: FrameworkAwareVisitor | None = None
) -> int | None:
    if not visitor:
        return None
    if definition.line in visitor.framework_decorated_lines:
        objects_with_routes = getattr(visitor, "objects_with_routes", None)
        if isinstance(objects_with_routes, dict) and objects_with_routes:
            for obj_name, route_lines in objects_with_routes.items():
                if definition.line in route_lines:
                    objects_passed = getattr(visitor, "objects_passed_as_args", set())
                    objects_created = getattr(visitor, "objects_created_by_call", set())
                    has_evidence = (
                        obj_name in objects_passed or obj_name in objects_created
                    )
                    if not has_evidence:
                        why_reduced = getattr(
                            definition, "why_confidence_reduced", None
                        )
                        if isinstance(why_reduced, list):
                            why_reduced.append("unregistered_router")
                        fw_signals = getattr(definition, "framework_signals", None)
                        if isinstance(fw_signals, list):
                            fw_signals.append(f"route_on_{obj_name}")
                        return 20
        return 0
    return None



# test_gate_kwargs.py
from skylos.gatekeeper import run_gate_interaction, check_gate
from skylos.constants import parse_exclude_folders


def test_gate_interaction_accepts_kwargs():
    result = {"unused_functions": [], "danger": [], "quality": [], "secrets": []}
    config = {}
    exit_code = run_gate_interaction(
        result=result, config=config, strict=False, force=False
    )
    # no findings = pass
    assert exit_code == 0


def test_gate_interaction_fails_on_positional():
    import pytest

    with pytest.raises(TypeError):
        run_gate_interaction({}, {})


def test_check_gate_strict_with_findings():
    result = {"unused_functions": [{"name": "foo", "file": "x.py", "line": 1}]}
    passed, reasons = check_gate(result, {}, strict=True)
    assert passed is False
    assert "Strict mode" in reasons[0]


def test_check_gate_passes_clean():
    result = {"unused_functions": [], "danger": [], "quality": [], "secrets": []}
    passed, reasons = check_gate(result, {})
    assert passed is True
    assert reasons == []


def test_default_excludes_filter_venv():
    excludes = parse_exclude_folders(use_defaults=True)
    assert "venv" in excludes or ".venv" in excludes
    assert "__pycache__" in excludes
    assert ".git" in excludes


def test_agent_review_respects_excludes(tmp_path):
    (tmp_path / "app.py").write_text("x = 1")
    venv = tmp_path / "venv" / "lib" / "pkg"
    venv.mkdir(parents=True)
    (venv / "big.py").write_text("y = 2")

    excludes = parse_exclude_folders(use_defaults=True)

    all_py = list(tmp_path.rglob("*.py"))
    filtered = [f for f in all_py if not any(ex in f.parts for ex in excludes)]

    assert len(all_py) == 2
    # venv one is excluded
    assert len(filtered) == 1
    assert filtered[0].name == "app.py"



from skylos.defend.plugin import DefensePlugin
from skylos.defend.result import DefenseResult
from skylos.discover.integration import LLMIntegration
from skylos.discover.graph import AIIntegrationGraph


class RagContextIsolationPlugin(DefensePlugin):
    id = "rag-context-isolation"
    name = "RAG Context Isolation"
    severity = "high"
    owasp_llm = "LLM01"
    description = (
        "Retrieved context in RAG pipelines must be wrapped in delimiters "
        "before injection into prompts to prevent indirect prompt injection"
    )
    remediation = (
        "Wrap retrieved documents in clear delimiters (e.g., XML tags, "
        "triple backticks) before inserting into the prompt template."
    )

    def applies_to(self, integration: LLMIntegration) -> bool:
        return integration.has_rag_context

    def check(
        self, integration: LLMIntegration, graph: AIIntegrationGraph
    ) -> DefenseResult:
        if integration.has_prompt_delimiter:
            return self._pass(
                integration,
                integration.location,
                "RAG context has prompt delimiters for isolation",
            )

        return self._fail(
            integration,
            integration.location,
            "RAG pipeline lacks context isolation delimiters",
        )



class TestLLMProviderResolution:
    def test_provider_inferred_from_claude_model(self):
        model = "claude-3-7-sonnet-latest"
        if "claude" in model.lower():
            provider = "anthropic"
        else:
            provider = "openai"
        assert provider == "anthropic"

    def test_provider_inferred_from_gpt_model(self):
        model = "gpt-4.1"
        if "claude" in model.lower():
            provider = "anthropic"
        else:
            provider = "openai"
        assert provider == "openai"

    def test_provider_inferred_from_local_model(self):
        model = "qwen2.5-coder:7b"
        if "claude" in model.lower():
            provider = "anthropic"
        else:
            provider = "openai"
        assert provider == "openai"


class TestProviderResolutionChain:
    def _resolve_provider(self, cli_provider, env_provider, model):
        return (
            cli_provider
            or env_provider
            or ("anthropic" if "claude" in model.lower() else "openai")
        )

    def test_cli_flag_takes_precedence(self):
        result = self._resolve_provider(
            cli_provider="openai", env_provider="anthropic", model="claude-3-opus"
        )
        assert result == "openai"

    def test_env_var_used_when_no_cli_flag(self):
        result = self._resolve_provider(
            cli_provider=None, env_provider="anthropic", model="gpt-4.1"
        )
        assert result == "anthropic"

    def test_model_inference_as_fallback(self):
        result = self._resolve_provider(
            cli_provider=None, env_provider=None, model="claude-3-sonnet"
        )
        assert result == "anthropic"


class TestBaseURLResolution:
    def _resolve_base_url(self, cli_base_url, env_skylos_url, env_openai_url):
        return cli_base_url or env_skylos_url or env_openai_url

    def test_cli_base_url_takes_precedence(self):
        result = self._resolve_base_url(
            cli_base_url="http://cli:8000/v1",
            env_skylos_url="http://skylos:8000/v1",
            env_openai_url="http://openai:8000/v1",
        )
        assert result == "http://cli:8000/v1"

    def test_skylos_env_used_when_no_cli(self):
        result = self._resolve_base_url(
            cli_base_url=None,
            env_skylos_url="http://skylos:8000/v1",
            env_openai_url="http://openai:8000/v1",
        )
        assert result == "http://skylos:8000/v1"

    def test_openai_env_as_fallback(self):
        result = self._resolve_base_url(
            cli_base_url=None,
            env_skylos_url=None,
            env_openai_url="http://openai:8000/v1",
        )
        assert result == "http://openai:8000/v1"

    def test_none_when_no_base_url_set(self):
        result = self._resolve_base_url(
            cli_base_url=None, env_skylos_url=None, env_openai_url=None
        )
        assert result is None


class TestLocalEndpointDetection:
    def _is_local_endpoint(self, provider, base_url):
        if provider != "openai" or not base_url:
            return False

        local_hosts = ["localhost", "127.0.0.1", "0.0.0.0"]
        for h in local_hosts:
            if h in base_url:
                return True
        return False

    def test_localhost_detected(self):
        assert self._is_local_endpoint("openai", "http://localhost:11434/v1") is True

    def test_127_0_0_1_detected(self):
        assert self._is_local_endpoint("openai", "http://127.0.0.1:1234/v1") is True

    def test_0_0_0_0_detected(self):
        assert self._is_local_endpoint("openai", "http://0.0.0.0:8000/v1") is True

    def test_remote_url_not_local(self):
        assert self._is_local_endpoint("openai", "https://api.openai.com/v1") is False

    def test_anthropic_provider_not_local(self):
        assert self._is_local_endpoint("anthropic", "http://localhost:8000/v1") is False

    def test_no_base_url_not_local(self):
        assert self._is_local_endpoint("openai", None) is False


class TestAPIKeyRequirement:
    def _needs_api_key_prompt(self, api_key, using_local):
        if not api_key and using_local:
            return False
        if not api_key:
            return True
        return False

    def test_no_key_remote_needs_prompt(self):
        assert self._needs_api_key_prompt(None, using_local=False) is True

    def test_no_key_local_no_prompt(self):
        assert self._needs_api_key_prompt(None, using_local=True) is False

    def test_has_key_no_prompt(self):
        assert self._needs_api_key_prompt("my-key", using_local=False) is False
        assert self._needs_api_key_prompt("my-key", using_local=True) is False


class TestCLIArgumentParsing:
    def test_provider_choices_valid(self):
        valid_providers = ["openai", "anthropic"]
        assert "openai" in valid_providers
        assert "anthropic" in valid_providers

    def test_default_model(self):
        default_model = "gpt-4.1"
        assert default_model == "gpt-4.1"


class TestEndToEndScenarios:
    def _simulate_cli_setup(
        self,
        model,
        cli_provider=None,
        cli_base_url=None,
        env_provider=None,
        env_base_url=None,
        env_api_key=None,
    ):
        provider = cli_provider
        if not provider:
            provider = env_provider
        if not provider:
            if "claude" in model.lower():
                provider = "anthropic"
            else:
                provider = "openai"

        base_url = cli_base_url
        if not base_url:
            base_url = env_base_url

        local_hosts = ["localhost", "127.0.0.1", "0.0.0.0"]
        is_local_host = False
        if base_url:
            for h in local_hosts:
                if h in base_url:
                    is_local_host = True
                    break

        using_local = bool(provider == "openai" and base_url and is_local_host)

        api_key = env_api_key
        if not api_key and using_local:
            api_key = ""

        needs_key_prompt = not api_key and not using_local

        return {
            "provider": provider,
            "base_url": base_url,
            "using_local": using_local,
            "api_key": api_key,
            "needs_key_prompt": needs_key_prompt,
        }

    def test_scenario_cloud_openai(self):
        result = self._simulate_cli_setup(model="gpt-4.1", env_api_key="sk-xxx")
        assert result["provider"] == "openai"
        assert result["base_url"] is None
        assert result["using_local"] is False
        assert result["needs_key_prompt"] is False

    def test_scenario_cloud_anthropic(self):
        result = self._simulate_cli_setup(
            model="claude-3-7-sonnet-latest", env_api_key="sk-ant-xxx"
        )
        assert result["provider"] == "anthropic"
        assert result["base_url"] is None
        assert result["using_local"] is False

    def test_scenario_local_ollama(self):
        result = self._simulate_cli_setup(
            model="qwen2.5-coder:7b",
            cli_provider="openai",
            cli_base_url="http://localhost:11434/v1",
        )
        assert result["provider"] == "openai"
        assert result["base_url"] == "http://localhost:11434/v1"
        assert result["using_local"] is True
        assert result["needs_key_prompt"] is False

    def test_scenario_local_lmstudio(self):
        result = self._simulate_cli_setup(
            model="mistral",
            cli_provider="openai",
            cli_base_url="http://localhost:1234/v1",
        )
        assert result["provider"] == "openai"
        assert result["base_url"] == "http://localhost:1234/v1"
        assert result["using_local"] is True
        assert result["needs_key_prompt"] is False

    def test_scenario_local_vllm(self):
        result = self._simulate_cli_setup(
            model="meta-llama/Llama-2-7b",
            cli_provider="openai",
            cli_base_url="http://127.0.0.1:8000/v1",
        )
        assert result["provider"] == "openai"
        assert result["using_local"] is True
        assert result["needs_key_prompt"] is False

    def test_scenario_env_config_only(self):
        result = self._simulate_cli_setup(
            model="qwen2.5-coder:7b",
            env_provider="openai",
            env_base_url="http://localhost:11434/v1",
        )
        assert result["provider"] == "openai"
        assert result["base_url"] == "http://localhost:11434/v1"
        assert result["using_local"] is True

    def test_scenario_missing_key_remote(self):
        result = self._simulate_cli_setup(model="gpt-4.1", env_api_key=None)
        assert result["needs_key_prompt"] is True

    def test_scenario_force_openai_for_claude_model(self):
        result = self._simulate_cli_setup(
            model="claude-compat-model",
            cli_provider="openai",
            cli_base_url="http://localhost:8000/v1",
        )
        assert result["provider"] == "openai"
        assert result["using_local"] is True


class TestEdgeCases:
    def test_case_insensitive_claude_detection(self):
        models = ["Claude-3", "CLAUDE-opus", "claude-sonnet", "ClAuDe-haiku"]
        for model in models:
            if "claude" in model.lower():
                provider = "anthropic"
            else:
                provider = "openai"
            assert provider == "anthropic", f"Failed for {model}"

    def test_claude_in_model_name_substring(self):
        model = "my-fine-tuned-claude-variant"
        if "claude" in model.lower():
            provider = "anthropic"
        else:
            provider = "openai"
        assert provider == "anthropic"

    def test_empty_string_env_vars_treated_as_unset(self):
        cli_provider = None
        env_provider = ""
        model = "gpt-4"

        provider = (
            cli_provider
            or env_provider
            or ("anthropic" if "claude" in model.lower() else "openai")
        )
        assert provider == "openai"


class TestIPv6LocalDetection:
    def _is_local_endpoint(self, provider, base_url):
        if provider != "openai" or not base_url:
            return False
        local_hosts = ["localhost", "127.0.0.1", "0.0.0.0", "::1", "[::1]"]
        for h in local_hosts:
            if h in base_url:
                return True
        return False


class TestRealWorldModelNames:
    def _resolve_provider(self, model, cli_provider=None):
        return cli_provider or ("anthropic" if "claude" in model.lower() else "openai")

    def test_openai_models(self):
        openai_models = [
            "gpt-4.1",
            "gpt-4-turbo",
            "gpt-3.5-turbo",
            "o1-preview",
            "o1-mini",
        ]
        for model in openai_models:
            assert self._resolve_provider(model) == "openai", f"Failed for {model}"

    def test_anthropic_models(self):
        anthropic_models = [
            "claude-3-7-sonnet-latest",
            "claude-3-opus-20240229",
            "claude-3-haiku-20240307",
            "claude-2.1",
        ]
        for model in anthropic_models:
            assert self._resolve_provider(model) == "anthropic", f"Failed for {model}"

    def test_local_models(self):
        local_models = [
            "qwen2.5-coder:7b",
            "llama3.2:latest",
            "mistral:7b-instruct",
            "codellama:13b",
            "deepseek-coder:6.7b",
            "phi3:mini",
        ]
        for model in local_models:
            assert self._resolve_provider(model) == "openai", f"Failed for {model}"


class TestPortVariations:
    def _simulate_cli_setup(self, base_url):
        local_hosts = ["localhost", "127.0.0.1", "0.0.0.0"]
        is_local = False
        if base_url:
            for h in local_hosts:
                if h in base_url:
                    is_local = True
                    break
        return is_local

    def test_ollama_default_port(self):
        assert self._simulate_cli_setup("http://localhost:11434/v1") is True

    def test_lmstudio_default_port(self):
        assert self._simulate_cli_setup("http://localhost:1234/v1") is True

    def test_vllm_default_port(self):
        assert self._simulate_cli_setup("http://localhost:8000/v1") is True

    def test_custom_port(self):
        assert self._simulate_cli_setup("http://localhost:9999/v1") is True

    def test_no_port_specified(self):
        assert self._simulate_cli_setup("http://localhost/v1") is True



"""Tests for community rules: taint-flow patterns, YAML loading, CLI commands."""

from __future__ import annotations

import ast
import os
import textwrap
from pathlib import Path
from unittest.mock import patch, MagicMock

import pytest
import yaml

from skylos.rules.custom import YAMLRule, load_custom_rules, load_community_rules
from skylos.linter import LinterVisitor


def _make_taint_rule(
    sources=None, sinks=None, sanitizers=None, rule_id="TEST-001"
):
    return YAMLRule(
        {
            "rule_id": rule_id,
            "name": "test taint rule",
            "severity": "HIGH",
            "category": "security",
            "yaml_config": {
                "message": "Tainted input flows to sink",
                "pattern": {
                    "type": "taint_flow",
                    "sources": sources or ["request.form", "request.args"],
                    "sinks": sinks or ["cursor.execute", "db.execute"],
                    "sanitizers": sanitizers or ["int", "escape_string"],
                },
            },
        }
    )


def _run_rule_on_code(rule, code):
    """Run a rule against source code and return findings."""
    tree = ast.parse(textwrap.dedent(code))
    linter = LinterVisitor([rule], "test.py")
    linter.visit(tree)
    return linter.findings


class TestTaintFlowDetection:
    """Tests for taint_flow pattern type in YAMLRule."""

    def test_source_to_sink_flagged(self):
        """Direct tainted variable used in sink call should be flagged."""
        code = """
        def handle():
            user_input = request.form["name"]
            cursor.execute(user_input)
        """
        rule = _make_taint_rule()
        findings = _run_rule_on_code(rule, code)
        assert len(findings) >= 1
        assert findings[0]["rule_id"] == "TEST-001"
        assert findings[0]["severity"] == "HIGH"

    def test_source_to_sink_via_fstring(self):
        """Tainted variable in f-string passed to sink should be flagged."""
        code = """
        def handle():
            name = request.args["name"]
            cursor.execute(f"SELECT * FROM users WHERE name = '{name}'")
        """
        rule = _make_taint_rule()
        findings = _run_rule_on_code(rule, code)
        assert len(findings) >= 1

    def test_source_to_sink_via_format(self):
        """Tainted variable in .format() passed to sink should be flagged."""
        code = """
        def handle():
            name = request.form["name"]
            query = "SELECT * FROM users WHERE name = '{}'".format(name)
            db.execute(query)
        """
        rule = _make_taint_rule()
        findings = _run_rule_on_code(rule, code)
        assert len(findings) >= 1

    def test_sanitizer_removes_taint(self):
        """Sanitizer call on tainted variable should remove taint."""
        code = """
        def handle():
            user_id = request.form["id"]
            safe_id = int(user_id)
            cursor.execute(safe_id)
        """
        rule = _make_taint_rule()
        findings = _run_rule_on_code(rule, code)
        # safe_id should NOT be tainted because int() is a sanitizer
        taint_findings = [f for f in findings if f["rule_id"] == "TEST-001"]
        assert len(taint_findings) == 0

    def test_no_source_no_finding(self):
        """Non-source data flowing to sink should not be flagged."""
        code = """
        def handle():
            data = "safe_static_string"
            cursor.execute(data)
        """
        rule = _make_taint_rule()
        findings = _run_rule_on_code(rule, code)
        assert len(findings) == 0

    def test_source_without_sink_no_finding(self):
        """Source data that never reaches a sink should not be flagged."""
        code = """
        def handle():
            user_input = request.form["name"]
            print(user_input)
        """
        rule = _make_taint_rule()
        findings = _run_rule_on_code(rule, code)
        assert len(findings) == 0

    def test_direct_source_in_sink_call(self):
        """Source expression directly as sink argument should be flagged."""
        code = """
        def handle():
            cursor.execute(request.form["query"])
        """
        rule = _make_taint_rule()
        findings = _run_rule_on_code(rule, code)
        assert len(findings) >= 1

    def test_function_scope_reset(self):
        """Taint state should reset at function boundaries."""
        code = """
        def first():
            user_input = request.form["name"]

        def second():
            cursor.execute(user_input)
        """
        rule = _make_taint_rule()
        findings = _run_rule_on_code(rule, code)
        # user_input in second() should not be tainted (different scope)
        assert len(findings) == 0

    def test_multiple_sinks(self):
        """Multiple sink calls with same tainted variable should produce multiple findings."""
        code = """
        def handle():
            name = request.form["name"]
            cursor.execute(name)
            db.execute(name)
        """
        rule = _make_taint_rule()
        findings = _run_rule_on_code(rule, code)
        assert len(findings) >= 2


class TestLoadCommunityRules:
    """Tests for load_community_rules function."""

    def test_load_from_directory(self, tmp_path):
        """Load rules from .yml files in a directory."""
        rules_dir = tmp_path / ".skylos" / "rules"
        rules_dir.mkdir(parents=True)

        rule_content = {
            "rules": [
                {
                    "id": "COMM-001",
                    "name": "Test rule",
                    "severity": "HIGH",
                    "category": "security",
                    "message": "Test message",
                    "pattern": {"type": "call", "function_match": ["eval"]},
                }
            ]
        }
        (rules_dir / "test_pack.yml").write_text(yaml.dump(rule_content))

        with patch("pathlib.Path.home", return_value=tmp_path):
            result = load_community_rules()

        assert len(result) == 1
        assert result[0]["rule_id"] == "COMM-001"
        assert result[0]["rule_type"] == "yaml"
        assert result[0]["enabled"] is True

    def test_load_empty_directory(self, tmp_path):
        """Return empty list when no rules installed."""
        rules_dir = tmp_path / ".skylos" / "rules"
        rules_dir.mkdir(parents=True)

        with patch("pathlib.Path.home", return_value=tmp_path):
            result = load_community_rules()

        assert result == []

    def test_load_no_directory(self, tmp_path):
        """Return empty list when rules directory doesn't exist."""
        with patch("pathlib.Path.home", return_value=tmp_path):
            result = load_community_rules()

        assert result == []

    def test_load_multiple_packs(self, tmp_path):
        """Load rules from multiple .yml files."""
        rules_dir = tmp_path / ".skylos" / "rules"
        rules_dir.mkdir(parents=True)

        for i in range(3):
            content = {
                "rules": [
                    {
                        "id": f"PACK{i}-001",
                        "name": f"Rule from pack {i}",
                        "severity": "MEDIUM",
                        "pattern": {"type": "call", "function_match": ["test"]},
                    }
                ]
            }
            (rules_dir / f"pack{i}.yml").write_text(yaml.dump(content))

        with patch("pathlib.Path.home", return_value=tmp_path):
            result = load_community_rules()

        assert len(result) == 3

    def test_load_skips_invalid_yaml(self, tmp_path):
        """Invalid YAML files should be silently skipped."""
        rules_dir = tmp_path / ".skylos" / "rules"
        rules_dir.mkdir(parents=True)

        (rules_dir / "bad.yml").write_text("{{{{not valid yaml")
        content = {
            "rules": [
                {
                    "id": "GOOD-001",
                    "name": "Good rule",
                    "severity": "LOW",
                    "pattern": {"type": "call", "function_match": ["test"]},
                }
            ]
        }
        (rules_dir / "good.yml").write_text(yaml.dump(content))

        with patch("pathlib.Path.home", return_value=tmp_path):
            result = load_community_rules()

        assert len(result) == 1
        assert result[0]["rule_id"] == "GOOD-001"

    def test_community_rules_become_yaml_rules(self, tmp_path):
        """Community rules should be loadable as YAMLRule objects."""
        rules_dir = tmp_path / ".skylos" / "rules"
        rules_dir.mkdir(parents=True)

        content = {
            "rules": [
                {
                    "id": "COMM-001",
                    "name": "Eval detection",
                    "severity": "HIGH",
                    "category": "security",
                    "message": "Avoid eval",
                    "pattern": {
                        "type": "call",
                        "function_match": ["eval"],
                        "args": {"is_dynamic": True, "position": 0},
                    },
                }
            ]
        }
        (rules_dir / "security.yml").write_text(yaml.dump(content))

        with patch("pathlib.Path.home", return_value=tmp_path):
            community_data = load_community_rules()

        rules = load_custom_rules(community_data)
        assert len(rules) == 1
        assert rules[0].rule_id == "COMM-001"
        assert isinstance(rules[0], YAMLRule)


class TestRuleValidation:
    """Tests for skylos rules validate logic."""

    def test_valid_rule_file(self, tmp_path):
        """A well-formed rule file should pass validation."""
        content = {
            "rules": [
                {
                    "id": "V-001",
                    "name": "Valid rule",
                    "severity": "HIGH",
                    "pattern": {"type": "call", "function_match": ["eval"]},
                }
            ]
        }
        rule_file = tmp_path / "valid.yml"
        rule_file.write_text(yaml.dump(content))

        data = yaml.safe_load(rule_file.read_text())
        assert "rules" in data
        rule = data["rules"][0]
        assert "id" in rule
        assert "name" in rule
        assert "severity" in rule
        assert "type" in rule["pattern"]

    def test_missing_id(self, tmp_path):
        """Rule missing 'id' should be caught by validation."""
        content = {
            "rules": [
                {
                    "name": "No ID rule",
                    "severity": "HIGH",
                    "pattern": {"type": "call"},
                }
            ]
        }
        rule_file = tmp_path / "bad.yml"
        rule_file.write_text(yaml.dump(content))

        data = yaml.safe_load(rule_file.read_text())
        rule = data["rules"][0]
        assert "id" not in rule

    def test_missing_pattern_type(self, tmp_path):
        """Rule with pattern missing 'type' should be caught."""
        content = {
            "rules": [
                {
                    "id": "V-002",
                    "name": "No type",
                    "severity": "HIGH",
                    "pattern": {"function_match": ["eval"]},
                }
            ]
        }
        rule_file = tmp_path / "notype.yml"
        rule_file.write_text(yaml.dump(content))

        data = yaml.safe_load(rule_file.read_text())
        assert "type" not in data["rules"][0]["pattern"]

    def test_taint_flow_requires_sources_and_sinks(self, tmp_path):
        """taint_flow rules must have sources and sinks."""
        content = {
            "rules": [
                {
                    "id": "V-003",
                    "name": "Bad taint",
                    "severity": "HIGH",
                    "pattern": {"type": "taint_flow"},
                }
            ]
        }
        rule_file = tmp_path / "badtaint.yml"
        rule_file.write_text(yaml.dump(content))

        data = yaml.safe_load(rule_file.read_text())
        pattern = data["rules"][0]["pattern"]
        assert pattern["type"] == "taint_flow"
        assert "sources" not in pattern
        assert "sinks" not in pattern

    def test_valid_taint_flow_rule(self, tmp_path):
        """A complete taint_flow rule should pass validation."""
        content = {
            "rules": [
                {
                    "id": "V-004",
                    "name": "SQL injection",
                    "severity": "HIGH",
                    "pattern": {
                        "type": "taint_flow",
                        "sources": ["request.form"],
                        "sinks": ["cursor.execute"],
                        "sanitizers": ["int"],
                    },
                }
            ]
        }
        rule_file = tmp_path / "taint.yml"
        rule_file.write_text(yaml.dump(content))

        data = yaml.safe_load(rule_file.read_text())
        pattern = data["rules"][0]["pattern"]
        assert pattern["type"] == "taint_flow"
        assert "sources" in pattern
        assert "sinks" in pattern


class TestRulesInstall:
    """Tests for rule installation with mocked filesystem."""

    def test_install_from_pack_name(self, tmp_path):
        """Install by pack name should download from known URL."""
        from skylos.cli import _rules_install

        rules_dir = tmp_path / "rules"
        console = MagicMock()

        yaml_content = yaml.dump(
            {
                "rules": [
                    {
                        "id": "PACK-001",
                        "name": "Pack rule",
                        "severity": "HIGH",
                        "pattern": {"type": "call", "function_match": ["eval"]},
                    }
                ]
            }
        )

        mock_resp = MagicMock()
        mock_resp.read.return_value = yaml_content.encode("utf-8")
        mock_resp.__enter__ = MagicMock(return_value=mock_resp)
        mock_resp.__exit__ = MagicMock(return_value=False)

        with patch("urllib.request.urlopen", return_value=mock_resp):
            _rules_install(console, rules_dir, "flask-security")

        installed = rules_dir / "flask-security.yml"
        assert installed.exists()
        data = yaml.safe_load(installed.read_text())
        assert len(data["rules"]) == 1

    def test_install_from_url(self, tmp_path):
        """Install by direct URL should use that URL."""
        from skylos.cli import _rules_install

        rules_dir = tmp_path / "rules"
        console = MagicMock()

        yaml_content = yaml.dump(
            {
                "rules": [
                    {
                        "id": "URL-001",
                        "name": "URL rule",
                        "severity": "MEDIUM",
                        "pattern": {"type": "call", "function_match": ["exec"]},
                    }
                ]
            }
        )

        mock_resp = MagicMock()
        mock_resp.read.return_value = yaml_content.encode("utf-8")
        mock_resp.__enter__ = MagicMock(return_value=mock_resp)
        mock_resp.__exit__ = MagicMock(return_value=False)

        with patch("urllib.request.urlopen", return_value=mock_resp):
            _rules_install(
                console,
                rules_dir,
                "https://example.com/my-rules.yml",
            )

        installed = rules_dir / "my-rules.yml"
        assert installed.exists()


class TestRulesRemove:
    """Tests for rule removal."""

    def test_remove_existing_pack(self, tmp_path):
        """Remove should delete the rule file."""
        from skylos.cli import _rules_remove

        rules_dir = tmp_path / "rules"
        rules_dir.mkdir(parents=True)
        (rules_dir / "mypack.yml").write_text("rules: []")

        console = MagicMock()
        _rules_remove(console, rules_dir, "mypack")

        assert not (rules_dir / "mypack.yml").exists()

    def test_remove_nonexistent_pack(self, tmp_path):
        """Remove of missing pack should exit with error."""
        from skylos.cli import _rules_remove

        rules_dir = tmp_path / "rules"
        rules_dir.mkdir(parents=True)
        console = MagicMock()

        with pytest.raises(SystemExit):
            _rules_remove(console, rules_dir, "nonexistent")



ENGINE_ID = "skylos-go"


def build_go_engine_args(engine_bin, root, skylos_version):
    return [
        engine_bin,
        "analyze",
        "--root",
        root,
        "--format",
        "json",
        "--skylos-version",
        skylos_version,
    ]


def validate_go_engine_output(obj):
    if type(obj) is not dict:
        raise ValueError("Go engine output must be a JSON object")

    if obj.get("engine") != ENGINE_ID:
        raise ValueError(
            "Go engine output has wrong engine id: %r" % (obj.get("engine"),)
        )

    version = obj.get("version")
    if type(version) is not str or not version.strip():
        raise ValueError("Go engine output missing/invalid version")

    findings = obj.get("findings")
    if type(findings) is not list:
        raise ValueError("Go engine output missing/invalid findings list")

    for f in findings:
        if type(f) is not dict:
            raise ValueError("Go engine findings must be objects")

    symbols = obj.get("symbols")
    if symbols is not None:
        if type(symbols) is not dict:
            raise ValueError("Go engine symbols must be a JSON object")
        if type(symbols.get("defs")) is not list:
            raise ValueError("Go engine symbols missing/invalid defs list")
        if type(symbols.get("refs")) is not list:
            raise ValueError("Go engine symbols missing/invalid refs list")

    return obj



import ast
from skylos.rules.quality.logic import UnusedExceptVarRule, ReturnConsistencyRule
from skylos.rules.quality.class_size import GodClassRule


def check_code(rule, code, filename="test.py"):
    tree = ast.parse(code)
    findings = []
    context = {"filename": filename, "mod": "test_module"}
    for node in ast.walk(tree):
        res = rule.visit_node(node, context)
        if res:
            findings.extend(res)
    return findings


class TestUnusedExceptVar:
    def test_unused_exception_variable(self):
        code = "try:\n    pass\nexcept ValueError as e:\n    print('oops')\n"
        findings = check_code(UnusedExceptVarRule(), code)
        assert len(findings) == 1
        assert findings[0]["rule_id"] == "SKY-L005"
        assert "e" in findings[0]["message"]

    def test_used_exception_variable(self):
        code = "try:\n    pass\nexcept ValueError as e:\n    print(e)\n"
        findings = check_code(UnusedExceptVarRule(), code)
        assert len(findings) == 0

    def test_bare_except_no_var(self):
        code = "try:\n    pass\nexcept:\n    pass\n"
        findings = check_code(UnusedExceptVarRule(), code)
        assert len(findings) == 0

    def test_underscore_convention(self):
        code = "try:\n    pass\nexcept ValueError as _:\n    print('ignored')\n"
        findings = check_code(UnusedExceptVarRule(), code)
        assert len(findings) == 1

    def test_multiple_except_one_unused(self):
        code = (
            "try:\n"
            "    pass\n"
            "except ValueError as e:\n"
            "    print('oops')\n"
            "except TypeError as e2:\n"
            "    print(e2)\n"
        )
        findings = check_code(UnusedExceptVarRule(), code)
        assert len(findings) == 1
        assert findings[0]["name"] == "e"

    def test_used_in_logging(self):
        code = (
            "import logging\n"
            "try:\n"
            "    pass\n"
            "except Exception as exc:\n"
            "    logging.error(exc)\n"
        )
        findings = check_code(UnusedExceptVarRule(), code)
        assert len(findings) == 0


class TestReturnConsistency:
    def test_inconsistent_explicit_return_vs_bare_return(self):
        code = "def f(x):\n    if x > 0:\n        return x * 2\n    return\n"
        findings = check_code(ReturnConsistencyRule(), code)
        assert len(findings) == 1
        assert findings[0]["rule_id"] == "SKY-L006"
        assert "inconsistent" in findings[0]["message"].lower()

    def test_consistent_return_value(self):
        code = "def f(x):\n    if x > 0:\n        return x * 2\n    return 0\n"
        findings = check_code(ReturnConsistencyRule(), code)
        assert len(findings) == 0

    def test_consistent_return_none(self):
        code = "def f(x):\n    if x > 0:\n        return\n    return\n"
        findings = check_code(ReturnConsistencyRule(), code)
        assert len(findings) == 0

    def test_explicit_return_none_mixed(self):
        code = "def f(x):\n    if x > 0:\n        return x\n    return None\n"
        findings = check_code(ReturnConsistencyRule(), code)
        assert len(findings) == 1

    def test_async_function_inconsistent(self):
        code = "async def f(x):\n    if x > 0:\n        return x\n    return\n"
        findings = check_code(ReturnConsistencyRule(), code)
        assert len(findings) == 1

    def test_nested_function_not_confused(self):
        code = (
            "def outer(x):\n"
            "    def inner():\n"
            "        return 42\n"
            "    if x:\n"
            "        return x\n"
            "    return 0\n"
        )
        findings = check_code(ReturnConsistencyRule(), code)
        assert len(findings) == 0

    def test_only_implicit_return_no_flag(self):
        code = "def f(x):\n    if x > 0:\n        return x * 2\n"
        findings = check_code(ReturnConsistencyRule(), code)
        assert len(findings) == 0


class TestGodClass:
    def _make_big_class(self, method_count, attr_count):
        methods = ""
        for i in range(method_count):
            methods += f"    def m{i}(self):\n        self.attr{i} = {i}\n"
        if attr_count > method_count:
            init_body = ""
            for i in range(method_count, attr_count):
                init_body += f"        self.attr{i} = {i}\n"
            methods = f"    def __init__(self):\n{init_body}" + methods
        return f"class BigClass:\n{methods}"

    def test_too_many_methods(self):
        code = self._make_big_class(method_count=21, attr_count=5)
        findings = check_code(GodClassRule(), code)
        rule_ids = [f["rule_id"] for f in findings]
        assert "SKY-Q501" in rule_ids
        method_finding = [
            f for f in findings if isinstance(f["value"], int) and f["value"] >= 21
        ][0]
        assert method_finding["threshold"] == 20

    def test_too_many_attributes(self):
        code = self._make_big_class(method_count=5, attr_count=16)
        findings = check_code(GodClassRule(), code)
        rule_ids = [f["rule_id"] for f in findings]
        assert "SKY-Q501" in rule_ids
        attr_finding = [f for f in findings if f["value"] == 16][0]
        assert attr_finding["threshold"] == 15

    def test_both_violations(self):
        code = self._make_big_class(method_count=21, attr_count=16)
        findings = check_code(GodClassRule(), code)
        assert len(findings) == 2

    def test_small_class_safe(self):
        code = (
            "class SmallClass:\n"
            "    def __init__(self):\n"
            "        self.x = 1\n"
            "    def do_thing(self):\n"
            "        pass\n"
        )
        findings = check_code(GodClassRule(), code)
        assert len(findings) == 0

    def test_custom_thresholds(self):
        code = self._make_big_class(method_count=6, attr_count=4)
        findings = check_code(GodClassRule(max_methods=5, max_attributes=3), code)
        method_findings = [
            f
            for f in findings
            if isinstance(f["value"], int) and f["value"] >= 6 and f["threshold"] == 5
        ]
        attr_findings = [
            f
            for f in findings
            if isinstance(f["value"], int) and f["value"] >= 4 and f["threshold"] == 3
        ]
        assert len(method_findings) >= 1
        assert len(attr_findings) >= 1



import ast
from skylos.rules.base import SkylosRule

DANGEROUS_CALLS = {
    "eval": ("SKY-D201", "HIGH", "Use of eval()"),
    "exec": ("SKY-D202", "HIGH", "Use of exec()"),
    "os.system": ("SKY-D203", "CRITICAL", "Use of os.system()"),
    "pickle.load": (
        "SKY-D204",
        "CRITICAL",
        "Untrusted deserialization via pickle.load",
    ),
    "pickle.loads": (
        "SKY-D205",
        "CRITICAL",
        "Untrusted deserialization via pickle.loads",
    ),
    "yaml.load": ("SKY-D206", "HIGH", "yaml.load without SafeLoader"),
    "hashlib.md5": ("SKY-D207", "MEDIUM", "Weak hash (MD5)"),
    "hashlib.sha1": ("SKY-D208", "MEDIUM", "Weak hash (SHA1)"),
    "subprocess.*": (
        "SKY-D209",
        "HIGH",
        "subprocess call with shell=True",
        {"kw_equals": {"shell": True}},
    ),
    "requests.*": (
        "SKY-D210",
        "HIGH",
        "requests call with verify=False",
        {"kw_equals": {"verify": False}},
    ),
    "marshal.loads": (
        "SKY-D233",
        "CRITICAL",
        "Untrusted deserialization via marshal.loads",
    ),
    "shelve.open": (
        "SKY-D233",
        "HIGH",
        "Untrusted deserialization via shelve.open",
    ),
    "jsonpickle.decode": (
        "SKY-D233",
        "CRITICAL",
        "Untrusted deserialization via jsonpickle.decode",
    ),
    "dill.loads": (
        "SKY-D233",
        "CRITICAL",
        "Untrusted deserialization via dill.loads",
    ),
    "dill.load": (
        "SKY-D233",
        "CRITICAL",
        "Untrusted deserialization via dill.load",
    ),
    ".exec_command": (
        "SKY-D235",
        "HIGH",
        "Remote command execution via exec_command (e.g., paramiko SSH)",
    ),
}


def _qualified_name_from_call(node: ast.Call):
    func = node.func
    parts = []
    while isinstance(func, ast.Attribute):
        parts.append(func.attr)
        func = func.value
    if isinstance(func, ast.Name):
        parts.append(func.id)
        parts.reverse()
        return ".".join(parts)
    return None


def _matches_rule(name, rule_key):
    if not name:
        return False
    if rule_key.startswith("."):
        return name.endswith(rule_key)
    if rule_key.endswith(".*"):
        return name.startswith(rule_key[:-2] + ".")
    return name == rule_key


def _kw_equals(node: ast.Call, requirements):
    if not requirements:
        return True
    kw_map = {}
    for kw in node.keywords or []:
        if kw.arg:
            if isinstance(kw.value, ast.Constant):
                kw_map[kw.arg] = kw.value.value

    for key, expected in requirements.items():
        val = kw_map.get(key)
        if val != expected:
            return False
    return True


def _yaml_load_without_safeloader(node: ast.Call):
    for kw in node.keywords or []:
        if kw.arg == "Loader":
            if "SafeLoader" in ast.dump(kw.value):
                return False
    return True


class DangerousCallsRule(SkylosRule):
    rule_id = "SKY-D200"
    name = "Dangerous Function Calls"

    def visit_node(self, node, context):
        if not isinstance(node, ast.Call):
            return None

        name = _qualified_name_from_call(node)
        if not name:
            return None

        findings = []

        for rule_key, tup in DANGEROUS_CALLS.items():
            if not _matches_rule(name, rule_key):
                continue

            rule_id = tup[0]
            severity = tup[1]
            message = tup[2]
            opts = tup[3] if len(tup) > 3 else None

            if rule_key == "yaml.load":
                if not _yaml_load_without_safeloader(node):
                    continue

            if opts and "kw_equals" in opts:
                if not _kw_equals(node, opts["kw_equals"]):
                    continue

            findings.append(
                {
                    "rule_id": rule_id,
                    "severity": severity,
                    "message": message,
                    "file": context.get("filename"),
                    "line": node.lineno,
                    "col": node.col_offset,
                }
            )
            break

        return findings if findings else None



from skylos.defend.plugin import DefensePlugin
from skylos.defend.result import DefenseResult
from skylos.discover.integration import LLMIntegration
from skylos.discover.graph import AIIntegrationGraph


class OutputPiiFilterPlugin(DefensePlugin):
    id = "output-pii-filter"
    name = "Output PII Filter"
    severity = "high"
    owasp_llm = "LLM06"
    description = (
        "LLM output returned to external users must pass through "
        "PII detection/redaction (presidio, scrubadub, or similar)"
    )
    remediation = (
        "Add PII detection and redaction between LLM response and user-facing output. "
        "Consider using presidio-analyzer, scrubadub, or a custom PII filter."
    )

    def applies_to(self, integration: LLMIntegration) -> bool:
        return bool(integration.input_sources)

    def check(
        self, integration: LLMIntegration, graph: AIIntegrationGraph
    ) -> DefenseResult:
        if integration.has_pii_filter:
            return self._pass(
                integration,
                integration.location,
                "PII filtering detected on LLM output path",
            )

        return self._fail(
            integration,
            integration.location,
            "No PII filtering on LLM output returned to users",
        )



import ast
import json
import tempfile
from pathlib import Path

from skylos.visitor import Visitor, Definition
from skylos.implicit_refs import ImplicitRefTracker


class TestGetAttrFStringPattern:
    def test_inline_fstring_registers_pattern(self):
        code = """
import sys

def export_csv(data):
    return ",".join(str(v) for v in data)

def export_json(data):
    import json
    return json.dumps(data)

def run_export(data, fmt):
    handler = getattr(sys.modules[__name__], f"export_{fmt}", None)
    return handler(data)
"""
        tracker = ImplicitRefTracker()
        v = Visitor("mymod", "test.py")
        v.pattern_tracker = tracker
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        patterns = [entry[2] for entry in tracker._compiled_patterns]
        assert any("export_" in p for p in patterns)

    def test_inline_fstring_marks_matching_defs(self):
        code = """
import sys

def export_csv(data):
    pass

def export_json(data):
    pass

def run_export(data, fmt):
    handler = getattr(sys.modules[__name__], f"export_{fmt}", None)
    return handler(data)
"""
        tracker = ImplicitRefTracker()
        v = Visitor("mymod", "test.py")
        v.pattern_tracker = tracker
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        for d in v.defs:
            if d.simple_name in ("export_csv", "export_json"):
                matched, _, _ = tracker.should_mark_as_used(d)
                assert matched, f"{d.simple_name} should match pattern"

    def test_non_matching_defs_not_marked(self):
        code = """
import sys

def export_csv(data):
    pass

def unrelated_func():
    pass

def run_export(data, fmt):
    handler = getattr(sys.modules[__name__], f"export_{fmt}", None)
    return handler(data)
"""
        tracker = ImplicitRefTracker()
        v = Visitor("mymod", "test.py")
        v.pattern_tracker = tracker
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        for d in v.defs:
            if d.simple_name == "unrelated_func":
                matched, _, _ = tracker.should_mark_as_used(d)
                assert not matched, "unrelated_func should not match export_* pattern"


class TestGlobalsFStringPattern:
    def test_globals_fstring_registers_pattern(self):
        code = """
def handle_create(payload):
    return {"action": "created"}

def handle_delete(payload):
    return {"action": "deleted"}

HANDLER_MAP = {a: globals()[f"handle_{a}"] for a in ("create", "delete")}
"""
        tracker = ImplicitRefTracker()
        v = Visitor("mymod", "test.py")
        v.pattern_tracker = tracker
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        patterns = [entry[2] for entry in tracker._compiled_patterns]
        assert any("handle_" in p for p in patterns)

    def test_globals_fstring_marks_matching_defs(self):
        code = """
def handle_create(payload):
    pass

def handle_update(payload):
    pass

HANDLER_MAP = {a: globals()[f"handle_{a}"] for a in ("create", "update")}
"""
        tracker = ImplicitRefTracker()
        v = Visitor("mymod", "test.py")
        v.pattern_tracker = tracker
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        for d in v.defs:
            if d.simple_name in ("handle_create", "handle_update"):
                matched, _, _ = tracker.should_mark_as_used(d)
                assert matched, f"{d.simple_name} should match pattern"


class TestInitSubclassRegistry:
    def test_base_classes_populated(self):
        code = """
class Base:
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

class Child(Base):
    pass
"""
        v = Visitor("mymod", "test.py")
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        child = None
        for d in v.defs:
            if d.simple_name == "Child":
                child = d
                break

        assert child is not None
        assert "mymod.Base" in child.base_classes

    def test_subclasses_not_flagged_unused(self):
        code = """
_REGISTRY = {}

class RegisteredHandler:
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        _REGISTRY[cls.name] = cls

class EmailHandler(RegisteredHandler):
    name = "email"

class SlackHandler(RegisteredHandler):
    name = "slack"

def get_handler(name):
    return _REGISTRY.get(name)
"""
        from skylos.analyzer import Skylos

        with tempfile.TemporaryDirectory() as tmpdir:
            p = Path(tmpdir) / "registry.py"
            p.write_text(code)

            s = Skylos()
            result = json.loads(s.analyze(tmpdir, thr=0))
            unused_classes = {
                item["simple_name"] for item in result.get("unused_classes", [])
            }

            assert "EmailHandler" not in unused_classes
            assert "SlackHandler" not in unused_classes

    def test_transitive_subclass_not_flagged_unused(self):
        """Subclasses via intermediate base should also be suppressed."""
        code = """
_REGISTRY = {}

class Base:
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        _REGISTRY[cls.name] = cls

class Intermediate(Base):
    pass

class LeafA(Intermediate):
    name = "a"

class LeafB(Intermediate):
    name = "b"

def get(name):
    return _REGISTRY.get(name)
"""
        from skylos.analyzer import Skylos

        with tempfile.TemporaryDirectory() as tmpdir:
            p = Path(tmpdir) / "registry.py"
            p.write_text(code)

            s = Skylos()
            result = json.loads(s.analyze(tmpdir, thr=0))
            unused_classes = {
                item["simple_name"] for item in result.get("unused_classes", [])
            }

            assert "LeafA" not in unused_classes
            assert "LeafB" not in unused_classes
            assert "Intermediate" not in unused_classes


class TestEnumMemberDetection:
    def _get_unused(self, code):
        from skylos.analyzer import Skylos

        with tempfile.TemporaryDirectory() as tmpdir:
            p = Path(tmpdir) / "mod.py"
            p.write_text(code)
            s = Skylos()
            result = json.loads(s.analyze(tmpdir, thr=0, grep_verify=False))
            unused = set()
            for cat in ["unused_functions", "unused_classes", "unused_variables"]:
                for item in result.get(cat, []):
                    unused.add(item["simple_name"])
            return unused

    def test_basic_enum_members_not_flagged(self):
        unused = self._get_unused("""
from enum import Enum

class Color(Enum):
    RED = 1
    GREEN = 2
    BLUE = 3

c = Color.RED
""")
        assert "RED" not in unused
        assert "GREEN" not in unused
        assert "BLUE" not in unused

    def test_int_enum_members_not_flagged(self):
        unused = self._get_unused("""
from enum import IntEnum

class Status(IntEnum):
    PENDING = 0
    ACTIVE = 1

s = Status.ACTIVE
""")
        assert "PENDING" not in unused
        assert "ACTIVE" not in unused

    def test_enum_methods_not_flagged(self):
        unused = self._get_unused("""
from enum import Enum

class Color(Enum):
    RED = 1
    GREEN = 2

    def hex_code(self):
        return f"#{self.value:06x}"

c = Color.RED
""")
        assert "hex_code" not in unused

    def test_indirect_enum_inheritance(self):
        unused = self._get_unused("""
from enum import Enum

class BaseEnum(Enum):
    pass

class Priority(BaseEnum):
    LOW = 1
    HIGH = 2

p = Priority.HIGH
""")
        assert "LOW" not in unused
        assert "HIGH" not in unused

    def test_flag_enum_members_not_flagged(self):
        unused = self._get_unused("""
from enum import Flag

class Permission(Flag):
    READ = 1
    WRITE = 2
    EXECUTE = 4

p = Permission.READ | Permission.WRITE
""")
        assert "READ" not in unused
        assert "WRITE" not in unused
        assert "EXECUTE" not in unused


class TestDefinitionReasonFields:
    def test_definition_has_new_fields(self):
        d = Definition("test.func", "function", "test.py", 1)
        assert d.heuristic_refs == {}
        assert d.dynamic_signals == []
        assert d.framework_signals == []
        assert d.why_unused == []
        assert d.why_confidence_reduced == []
        assert d._attr_name_ref_count == 0

    def test_to_dict_includes_new_fields_when_populated(self):
        d = Definition("test.func", "function", "test.py", 1)
        d.heuristic_refs = {"same_file_attr": 1.0}
        d.dynamic_signals = ["inspect_getmembers"]
        result = d.to_dict()
        assert "heuristic_refs" in result
        assert result["heuristic_refs"] == {"same_file_attr": 1.0}
        assert "dynamic_signals" in result

    def test_to_dict_excludes_empty_fields(self):
        d = Definition("test.func", "function", "test.py", 1)
        result = d.to_dict()
        assert "heuristic_refs" not in result
        assert "dynamic_signals" not in result
        assert "why_unused" not in result


class TestContextTracking:
    def test_attr_context_tracked(self):
        code = """
class Foo:
    def run(self):
        pass

obj.run()
"""
        v = Visitor("mymod", "test.py")
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        assert "run" in v._used_attr_names
        contexts = [c for c in v._used_attr_names_with_context if c[0] == "run"]
        assert len(contexts) > 0
        assert contexts[0][1] == "mymod"


class TestLocalsHandling:
    def test_locals_subscript_adds_ref(self):
        code = """
def my_func():
    pass

def dispatch():
    f = locals()["my_func"]
    f()
"""
        v = Visitor("mymod", "test.py")
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        refs = [r[0] for r in v.refs]
        assert "my_func" in refs or "mymod.my_func" in refs


class TestVarsAndDictHandling:
    """Phase 2B: Test vars() and __dict__ subscript handling."""

    def test_vars_subscript_adds_ref(self):
        code = """
class Foo:
    def __init__(self):
        self.name = "test"

    def get_field(self, key):
        return vars(self)["name"]
"""
        v = Visitor("mymod", "test.py")
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        refs = [r[0] for r in v.refs]
        assert "name" in refs or "mymod.Foo.name" in refs

    def test_dict_subscript_adds_ref(self):
        code = """
class Bar:
    x = 10

    def get_x(self):
        return self.__dict__["x"]
"""
        v = Visitor("mymod", "test.py")
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        refs = [r[0] for r in v.refs]
        assert "x" in refs or "mymod.Bar.x" in refs


class TestRightSideConcat:
    def test_right_side_concat_pattern(self):
        code = """
def load_handler():
    name = "handler"
    handler = getattr(module, name + "_impl")
"""
        tracker = ImplicitRefTracker()
        v = Visitor("mymod", "test.py")
        v.pattern_tracker = tracker
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        patterns = [entry[2] for entry in tracker._compiled_patterns]
        assert any("_impl" in p for p in patterns)


class TestFormatInGetattr:
    def test_format_pattern(self):
        code = """
def dispatch(action):
    handler = getattr(module, "handle_{}".format(action))
"""
        tracker = ImplicitRefTracker()
        v = Visitor("mymod", "test.py")
        v.pattern_tracker = tracker
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        patterns = [entry[2] for entry in tracker._compiled_patterns]
        assert any("handle_" in p for p in patterns)


class TestInspectGetmembers:
    def test_inspect_getmembers_self_marks_class_members(self):
        code = """
import inspect

class Plugin:
    def action_one(self):
        pass

    def action_two(self):
        pass

    def get_actions(self):
        return inspect.getmembers(self)
"""
        v = Visitor("mymod", "test.py")
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        action_defs = [d for d in v.defs if d.simple_name.startswith("action_")]
        for d in action_defs:
            assert "inspect_getmembers" in d.dynamic_signals, (
                f"{d.simple_name} should have inspect_getmembers signal"
            )

    def test_inspect_getmembers_class_name(self):
        code = """
import inspect

class Handler:
    def process(self):
        pass

members = inspect.getmembers(Handler)
"""
        v = Visitor("mymod", "test.py")
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        process_def = [d for d in v.defs if d.simple_name == "process"]
        assert len(process_def) == 1
        assert "inspect_getmembers" in process_def[0].dynamic_signals


class TestDirDetection:
    def test_dir_self_marks_class_members(self):
        code = """
class Validator:
    def check_email(self):
        pass

    def check_name(self):
        pass

    def run_checks(self):
        for name in dir(self):
            if name.startswith("check_"):
                getattr(self, name)()
"""
        v = Visitor("mymod", "test.py")
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        check_defs = [d for d in v.defs if d.simple_name.startswith("check_")]
        for d in check_defs:
            assert "dir_self" in d.dynamic_signals, (
                f"{d.simple_name} should have dir_self signal"
            )

    def test_dir_class_name(self):
        code = """
class Config:
    debug = True
    verbose = False

attrs = dir(Config)
"""
        v = Visitor("mymod", "test.py")
        tree = ast.parse(code)
        v.visit(tree)
        v.finalize()

        config_members = [d for d in v.defs if "Config." in d.name]
        for d in config_members:
            assert "dir_class" in d.dynamic_signals


class TestWeightedHeuristicRefs:
    def _get_result(self, code, filename="mod.py"):
        from skylos.analyzer import Skylos

        with tempfile.TemporaryDirectory() as tmpdir:
            p = Path(tmpdir) / filename
            p.write_text(code)
            s = Skylos()
            return json.loads(s.analyze(tmpdir, thr=0))

    def test_heuristic_refs_in_json_output(self):
        code = """
class Service:
    def run(self):
        pass

obj.run()
"""
        result = self._get_result(code)
        all_items = []
        for cat in ["unused_functions", "unused_classes", "unused_variables"]:
            all_items.extend(result.get(cat, []))

        unused_names = {item["simple_name"] for item in all_items}
        assert "run" not in unused_names


class TestGetAttrScopeReduction:
    def _get_unused(self, files_dict):
        from skylos.analyzer import Skylos

        with tempfile.TemporaryDirectory() as tmpdir:
            for fname, code in files_dict.items():
                p = Path(tmpdir) / fname
                p.parent.mkdir(parents=True, exist_ok=True)
                p.write_text(code)
            s = Skylos()
            result = json.loads(s.analyze(tmpdir, thr=0, grep_verify=False))
            unused = set()
            for cat in ["unused_functions", "unused_classes", "unused_variables"]:
                for item in result.get(cat, []):
                    unused.add(item["simple_name"])
            return unused

    def test_common_name_cross_package_gets_lower_confidence(self):
        files = {
            "pkg/__init__.py": "",
            "pkg/service.py": """
class MyService:
    def get(self):
        '''This should be detected as unused'''
        return None
""",
            "pkg/main.py": """
data = {"key": "value"}
result = data.get("key")
""",
        }
        unused = self._get_unused(files)


class TestModuleReachabilityRefinement:
    def test_refine_removes_deep_modules(self):
        from skylos.module_reachability import ModuleReachabilityAnalyzer

        analyzer = ModuleReachabilityAnalyzer()
        analyzer.all_modules = {
            "pkg",
            "pkg.sub1",
            "pkg.sub2",
            "pkg.sub2.deep1",
            "pkg.sub2.deep2",
        }
        analyzer._getattr_packages = {"pkg"}
        analyzer.graph["pkg.sub1"].add("pkg.sub2.deep1")

        analyzer._expand_getattr_packages()

        assert "pkg.sub1" in analyzer.graph["pkg"]
        assert "pkg.sub2" in analyzer.graph["pkg"]
        assert "pkg.sub2.deep2" not in analyzer.graph["pkg"]


class TestUnderscoreVarargSuppression:
    """Regression: *_args and **_kwargs should not be flagged as unused."""

    def _get_unused_params(self, code):
        from skylos.analyzer import Skylos

        with tempfile.TemporaryDirectory() as tmpdir:
            p = Path(tmpdir) / "mod.py"
            p.write_text(code)
            s = Skylos()
            result = json.loads(s.analyze(tmpdir, thr=0))
            return {item["simple_name"] for item in result.get("unused_parameters", [])}

    def test_underscore_args_not_flagged(self):
        code = """
def fail_render(*_args, **_kwargs):
    raise AssertionError("should not be called")

fail_render()
"""
        unused = self._get_unused_params(code)
        assert "_args" not in unused
        assert "_kwargs" not in unused

    def test_bare_underscore_vararg_not_flagged(self):
        code = """
def noop(*_):
    pass

noop()
"""
        unused = self._get_unused_params(code)
        assert "_" not in unused

    def test_regular_args_still_flagged(self):
        code = """
def func(used, unused_param):
    return used

func(1, 2)
"""
        unused = self._get_unused_params(code)
        assert "unused_param" in unused


class TestSameFileVariableUsage:
    def _get_unused(self, code):
        from skylos.analyzer import Skylos

        with tempfile.TemporaryDirectory() as tmpdir:
            p = Path(tmpdir) / "mod.py"
            p.write_text(code)
            s = Skylos()
            result = json.loads(s.analyze(tmpdir, thr=0, grep_verify=False))
            unused = set()
            for cat in ["unused_functions", "unused_classes", "unused_variables"]:
                for item in result.get(cat, []):
                    unused.add(item["simple_name"])
            return unused

    def test_variable_in_fstring_not_flagged(self):
        code = """
BASE_URL = "https://example.com/"

def build_url(slug):
    return f"{BASE_URL}{slug}/"

build_url("test")
"""
        unused = self._get_unused(code)
        assert "BASE_URL" not in unused

    def test_string_only_mention_does_not_suppress(self):
        """A name appearing only inside a string literal must NOT keep it alive."""
        code = """
def entry():
    pass

NAME = "entry"
"""
        unused = self._get_unused(code)
        assert "entry" in unused


class TestRelativeImportResolution:
    def _get_unused(self, files_dict):
        from skylos.analyzer import Skylos

        with tempfile.TemporaryDirectory() as tmpdir:
            for fname, code in files_dict.items():
                p = Path(tmpdir) / fname
                p.parent.mkdir(parents=True, exist_ok=True)
                p.write_text(code)
            s = Skylos()
            result = json.loads(s.analyze(tmpdir, thr=0, grep_verify=False))
            unused = set()
            for cat in ["unused_functions", "unused_classes", "unused_variables"]:
                for item in result.get(cat, []):
                    unused.add(item["simple_name"])
            return unused

    def test_relative_import_keeps_constant_alive(self):
        files = {
            "pkg/__init__.py": "",
            "pkg/rules.py": 'BASE_URL = "https://example.com/"\n',
            "pkg/main.py": """
from .rules import BASE_URL

def build_url(slug):
    return f"{BASE_URL}{slug}/"

build_url("test")
""",
        }
        unused = self._get_unused(files)
        assert "BASE_URL" not in unused



from __future__ import annotations
import ast
import sys
from skylos.rules.danger.taint import TaintVisitor


REDIRECT_FUNCS = {"redirect", "HttpResponseRedirect", "HttpResponsePermanentRedirect"}

REQUEST_ARGS_ATTRS = {"args", "params", "query", "GET"}


def _is_request_args_get(node):
    if not isinstance(node, ast.Call):
        return False
    func = node.func
    if not isinstance(func, ast.Attribute) or func.attr != "get":
        return False
    val = func.value
    if (
        isinstance(val, ast.Attribute)
        and val.attr in REQUEST_ARGS_ATTRS
        and isinstance(val.value, ast.Name)
        and val.value.id == "request"
    ):
        return True
    return False


def _has_url_guard(func_body):
    for node in ast.walk(func_body):
        if (
            isinstance(node, ast.Attribute)
            and node.attr == "netloc"
            and isinstance(node.value, ast.Call)
        ):
            call_func = node.value.func
            name = None
            if isinstance(call_func, ast.Name):
                name = call_func.id
            elif isinstance(call_func, ast.Attribute):
                name = call_func.attr
            if name == "urlparse":
                return True
        if (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Attribute)
            and node.func.attr == "startswith"
            and node.args
            and isinstance(node.args[0], ast.Constant)
            and node.args[0].value == "/"
        ):
            return True
    return False


class _RedirectFlowChecker(TaintVisitor):
    def __init__(self, file_path, findings):
        super().__init__(file_path, findings)
        self._func_node_stack = []

    def visit_FunctionDef(self, node):
        self._func_node_stack.append(node)
        super().visit_FunctionDef(node)
        self._func_node_stack.pop()

    def visit_AsyncFunctionDef(self, node):
        self._func_node_stack.append(node)
        super().visit_AsyncFunctionDef(node)
        self._func_node_stack.pop()

    def visit_Call(self, node):
        func = node.func
        func_name = None

        if isinstance(func, ast.Name):
            func_name = func.id
        elif isinstance(func, ast.Attribute):
            func_name = func.attr

        if func_name in REDIRECT_FUNCS and node.args:
            url_arg = node.args[0]

            if _is_request_args_get(url_arg):
                enclosing = self._func_node_stack[-1] if self._func_node_stack else None
                if enclosing is None or not _has_url_guard(enclosing):
                    self.findings.append(
                        {
                            "rule_id": "SKY-D230",
                            "severity": "HIGH",
                            "message": (
                                "Possible open redirect: request parameter passed to "
                                "redirect via .get() — default value does not prevent "
                                "attacker-supplied input."
                            ),
                            "file": str(self.file_path),
                            "line": node.lineno,
                            "col": node.col_offset,
                            "symbol": self._current_symbol(),
                        }
                    )
            elif self.is_tainted(url_arg):
                self.findings.append(
                    {
                        "rule_id": "SKY-D230",
                        "severity": "HIGH",
                        "message": "Possible open redirect: user-controlled URL passed to redirect.",
                        "file": str(self.file_path),
                        "line": node.lineno,
                        "col": node.col_offset,
                        "symbol": self._current_symbol(),
                    }
                )

        self.generic_visit(node)


def scan(tree, file_path, findings):
    try:
        checker = _RedirectFlowChecker(file_path, findings)
        checker.visit(tree)
    except Exception as e:
        print(f"Redirect flow analysis failed for {file_path}: {e}", file=sys.stderr)



import os
from urllib.parse import urlparse

from skylos.credentials import get_key, save_key, PROVIDERS


LOCAL_HOSTS = {"localhost", "127.0.0.1", "0.0.0.0"}


def detect_provider(model):
    m = (model or "").strip().lower()

    for prefix in [
        "ollama/",
        "groq/",
        "gemini/",
        "mistral/",
        "together/",
        "deepseek/",
        "xai/",
    ]:
        if m.startswith(prefix):
            return prefix[:-1]

    if "claude" in m:
        return "anthropic"

    return "openai"


def is_local_llm(model, base_url: None):
    m = (model or "").strip().lower()
    if m.startswith("ollama/"):
        return True

    url = (base_url or "").strip()
    if not url:
        return False

    try:
        host = urlparse(url).hostname or ""
    except Exception:
        host = ""

    host = host.lower()
    return host in LOCAL_HOSTS


def resolve_llm_runtime(
    *,
    model,
    provider_override: None,
    base_url_override: None,
    console=None,
    allow_prompt=True,
):
    provider = (
        (provider_override or "").strip().lower()
        or (os.getenv("SKYLOS_LLM_PROVIDER") or "").strip().lower()
        or detect_provider(model)
    )

    base_url = (
        (base_url_override or "").strip()
        or (os.getenv("SKYLOS_LLM_BASE_URL") or "").strip()
        or (os.getenv("OPENAI_BASE_URL") or "").strip()
        or None
    )

    local = is_local_llm(model, base_url)

    if local or provider == "ollama":
        return provider, "", base_url, True

    api_key = get_key(provider)

    if api_key:
        env_var = PROVIDERS.get(provider)
        if env_var:
            os.environ.setdefault(env_var, api_key)

    if (not api_key) and allow_prompt and console is not None:
        env_var = PROVIDERS.get(provider)
        label = env_var or f"{provider.upper()}_API_KEY"
        console.print(f"[warn]No {label} found.[/warn]")
        try:
            api_key = console.input(
                f"[bold yellow]Paste {provider.title()} API Key:[/bold yellow] ",
                password=True,
            )
            if api_key:
                save_key(provider, api_key)
        except KeyboardInterrupt:
            return provider, None, base_url, False

    return provider, api_key, base_url, False



import os
import tempfile
import unittest
from pathlib import Path
from skylos.analyzer import Skylos


def _posix(p):
    return str(p).replace(os.sep, "/")


class TestPathExclusion(unittest.TestCase):
    def setUp(self):
        self.analyzer = Skylos()
        self.root = Path(".")

    def test_exclude_nested_folder_regression(self):
        excludes = ["src/legacy"]

        file_path = Path("src/legacy/old_main.py")
        self.assertTrue(
            self.analyzer._should_exclude_file(file_path, self.root, excludes),
            "Failed to exclude file inside nested folder 'src/legacy'",
        )

        deep_file = Path("src/legacy/utils/helpers.py")
        self.assertTrue(
            self.analyzer._should_exclude_file(deep_file, self.root, excludes),
            "Failed to exclude file deeply nested inside 'src/legacy'",
        )

    def test_exclude_nested_folder_false_positives(self):
        excludes = ["src/legacy"]

        similar_path = Path("src/legacy_v2/new.py")
        self.assertFalse(
            self.analyzer._should_exclude_file(similar_path, self.root, excludes),
            "Incorrectly excluded a folder sharing a prefix ('src/legacy_v2')",
        )

        diff_path = Path("src/modern/main.py")
        self.assertFalse(
            self.analyzer._should_exclude_file(diff_path, self.root, excludes),
            "Incorrectly excluded a non-matching sibling folder",
        )

    def test_windows_path_normalization(self):
        excludes = ["src/legacy"]

        file_path = Path("src") / "legacy" / "windows.py"

        self.assertTrue(
            self.analyzer._should_exclude_file(file_path, self.root, excludes),
            "Failed to exclude path constructed with OS separators",
        )

    def test_existing_wildcard_logic(self):
        excludes = ["*pycache__"]
        file_path = Path("src/__pycache__/cache.py")

        self.assertTrue(
            self.analyzer._should_exclude_file(file_path, self.root, excludes),
            "Regression: Wildcard exclusion failed",
        )

    def test_multiple_excludes_logic(self):
        excludes = ["src/legacy", "venv"]

        file_path = Path("venv/lib/site-packages/pkg.py")

        self.assertTrue(
            self.analyzer._should_exclude_file(file_path, self.root, excludes),
            "Regression: Loop stopped prematurely (did not check second exclude item)",
        )

    def test_trailing_slash_stripped(self):
        excludes = ["src/legacy/"]

        file_path = Path("src") / "legacy" / "old_main.py"
        self.assertTrue(
            self.analyzer._should_exclude_file(file_path, self.root, excludes),
            "Failed to exclude file when exclude path has trailing slash",
        )

        deep_file = Path("src") / "legacy" / "utils" / "helpers.py"
        self.assertTrue(
            self.analyzer._should_exclude_file(deep_file, self.root, excludes),
            "Failed to exclude deeply nested file when exclude path has trailing slash",
        )

    def test_trailing_slash_no_false_positive(self):
        excludes = ["src/legacy/"]

        similar_path = Path("src") / "legacy_v2" / "new.py"
        self.assertFalse(
            self.analyzer._should_exclude_file(similar_path, self.root, excludes),
            "Trailing-slash stripping widened the match to a prefix sibling",
        )

    def test_cwd_relative_exclude_with_target_prefix(self):
        tmpdir = tempfile.mkdtemp()
        try:
            versions_dir = Path(tmpdir) / "app" / "alembic" / "versions"
            versions_dir.mkdir(parents=True)
            (versions_dir / "001.py").touch()
            (Path(tmpdir) / "app" / "alembic" / "env.py").touch()

            ok_dir = Path(tmpdir) / "app" / "models"
            ok_dir.mkdir(parents=True, exist_ok=True)
            (ok_dir / "user.py").touch()

            target_root = Path(tmpdir) / "app"
            excludes = ["app/alembic"]

            self.assertTrue(
                self.analyzer._should_exclude_file(
                    target_root / "alembic" / "versions" / "001.py",
                    target_root,
                    excludes,
                ),
                f"app/alembic exclude failed on nested file "
                f"(rel={_posix(Path('alembic/versions/001.py'))})",
            )
            self.assertTrue(
                self.analyzer._should_exclude_file(
                    target_root / "alembic" / "env.py",
                    target_root,
                    excludes,
                ),
                f"app/alembic exclude failed on direct child "
                f"(rel={_posix(Path('alembic/env.py'))})",
            )
            self.assertFalse(
                self.analyzer._should_exclude_file(
                    target_root / "models" / "user.py",
                    target_root,
                    excludes,
                ),
                "app/alembic exclude incorrectly excluded models/user.py",
            )
        finally:
            import shutil

            shutil.rmtree(tmpdir, ignore_errors=True)

    def test_cwd_relative_exclude_with_trailing_slash(self):
        tmpdir = tempfile.mkdtemp()
        try:
            versions_dir = Path(tmpdir) / "app" / "alembic" / "versions"
            versions_dir.mkdir(parents=True)
            (versions_dir / "001_init.py").touch()
            (Path(tmpdir) / "app" / "alembic" / "env.py").touch()
            (Path(tmpdir) / "app" / "alembic" / "__init__.py").touch()

            ok_dir = Path(tmpdir) / "app" / "models"
            ok_dir.mkdir(parents=True, exist_ok=True)
            (ok_dir / "user.py").touch()

            target_root = Path(tmpdir) / "app"
            excludes = ["app/alembic/"]

            for rel in [
                Path("alembic") / "versions" / "001_init.py",
                Path("alembic") / "env.py",
                Path("alembic") / "__init__.py",
            ]:
                self.assertTrue(
                    self.analyzer._should_exclude_file(
                        target_root / rel,
                        target_root,
                        excludes,
                    ),
                    f"app/alembic/ exclude failed on {_posix(rel)}",
                )

            self.assertFalse(
                self.analyzer._should_exclude_file(
                    target_root / "models" / "user.py",
                    target_root,
                    excludes,
                ),
                "app/alembic/ exclude incorrectly excluded models/user.py",
            )
        finally:
            import shutil

            shutil.rmtree(tmpdir, ignore_errors=True)

    def test_exclude_normalization_uses_forward_slashes(self):
        excludes = ["src/legacy"]
        file_path = Path("src") / "legacy" / "deep" / "file.py"
        self.assertTrue(
            self.analyzer._should_exclude_file(file_path, self.root, excludes),
            f"Exclude failed after Path normalization: {_posix(file_path)}",
        )


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations
import ast
import sys
from skylos.rules.danger.taint import TaintVisitor, CMD_SANITIZERS


def _qualified_name(node):
    func = node.func
    parts = []
    while isinstance(func, ast.Attribute):
        parts.append(func.attr)
        func = func.value

    if isinstance(func, ast.Name):
        parts.append(func.id)
        parts.reverse()
        return ".".join(parts)

    return None


class _CmdFlowChecker(TaintVisitor):
    OS_SYSTEM = "os.system"
    SUBPROC_PREFIX = "subprocess."

    def visit_Call(self, node):
        qn = _qualified_name(node)
        if not qn:
            self.generic_visit(node)
            return

        if qn == self.OS_SYSTEM and node.args:
            if self.is_tainted(node.args[0]):
                self.findings.append(
                    {
                        "rule_id": "SKY-D212",
                        "severity": "CRITICAL",
                        "message": "Possible command injection (os.system): tainted input.",
                        "file": str(self.file_path),
                        "line": node.lineno,
                        "col": node.col_offset,
                        "symbol": self._current_symbol(),
                    }
                )

        if qn.endswith(".exec_command") and node.args:
            if self.is_tainted(node.args[0]):
                self.findings.append(
                    {
                        "rule_id": "SKY-D212",
                        "severity": "CRITICAL",
                        "message": "Possible command injection (ssh exec_command): tainted input.",
                        "file": str(self.file_path),
                        "line": node.lineno,
                        "col": node.col_offset,
                        "symbol": self._current_symbol(),
                    }
                )

        if qn.startswith(self.SUBPROC_PREFIX):
            shell_true = False
            for kw in node.keywords:
                if (
                    kw.arg == "shell"
                    and isinstance(kw.value, ast.Constant)
                    and kw.value.value is True
                ):
                    shell_true = True

            if shell_true and self.is_tainted(node):
                self.findings.append(
                    {
                        "rule_id": "SKY-D212",
                        "severity": "CRITICAL",
                        "message": "Possible command injection (subprocess shell=True): tainted input.",
                        "file": str(self.file_path),
                        "line": node.lineno,
                        "col": node.col_offset,
                    }
                )

        self.generic_visit(node)


def scan(tree, file_path, findings):
    try:
        checker = _CmdFlowChecker(file_path, findings, sanitizers=CMD_SANITIZERS)
        checker.visit(tree)
    except Exception as e:
        print(f"CMD flow failed for {file_path}: {e}", file=sys.stderr)



class TestModuleComputation:
    def _make_analyzer(self):
        from skylos.analyzer import Skylos

        return Skylos()

    def test_regular_layout(self, tmp_path):
        analyzer = self._make_analyzer()

        pkg = tmp_path / "mypackage"
        pkg.mkdir()
        (pkg / "__init__.py").touch()
        cli = pkg / "cli.py"
        cli.touch()

        result = analyzer._module(tmp_path, cli)
        assert result == "mypackage.cli"

    def test_src_layout(self, tmp_path):
        analyzer = self._make_analyzer()

        src = tmp_path / "src"
        src.mkdir()
        pkg = src / "mypackage"
        pkg.mkdir()
        (pkg / "__init__.py").touch()
        cli = pkg / "cli.py"
        cli.touch()

        result = analyzer._module(tmp_path, cli)
        assert result == "mypackage.cli"

    def test_monorepo_src_layout(self, tmp_path):
        analyzer = self._make_analyzer()

        pkg1_dir = tmp_path / "package1" / "src" / "pkg1"
        pkg1_dir.mkdir(parents=True)
        (pkg1_dir / "__init__.py").touch()
        cli = pkg1_dir / "cli.py"
        cli.touch()

        result = analyzer._module(tmp_path, cli)
        assert result == "pkg1.cli"

    def test_src_as_real_package(self, tmp_path):
        analyzer = self._make_analyzer()

        src = tmp_path / "src"
        src.mkdir()
        (src / "__init__.py").touch()
        cli = src / "cli.py"
        cli.touch()

        result = analyzer._module(tmp_path, cli)
        assert result == "src.cli"

    def test_nested_src_layout(self, tmp_path):
        analyzer = self._make_analyzer()

        nested = tmp_path / "sub" / "src" / "pkg"
        nested.mkdir(parents=True)
        (nested / "__init__.py").touch()
        mod = nested / "module.py"
        mod.touch()

        result = analyzer._module(tmp_path, mod)
        assert result == "pkg.module"

    def test_init_file(self, tmp_path):
        analyzer = self._make_analyzer()

        src = tmp_path / "src"
        src.mkdir()
        pkg = src / "mypackage"
        pkg.mkdir()
        init = pkg / "__init__.py"
        init.touch()

        result = analyzer._module(tmp_path, init)
        assert result == "mypackage"

    def test_deep_module(self, tmp_path):
        analyzer = self._make_analyzer()

        deep = tmp_path / "src" / "pkg" / "sub" / "deep"
        deep.mkdir(parents=True)
        mod = deep / "module.py"
        mod.touch()

        result = analyzer._module(tmp_path, mod)
        assert result == "pkg.sub.deep.module"



from __future__ import annotations
import ast
import sys


def _qualified_name(node):
    func = node.func
    parts = []
    while isinstance(func, ast.Attribute):
        parts.append(func.attr)
        func = func.value
    if isinstance(func, ast.Name):
        parts.append(func.id)
        parts.reverse()
        return ".".join(parts)
    return None


class _CORSChecker(ast.NodeVisitor):
    def __init__(self, file_path, findings):
        self.file_path = file_path
        self.findings = findings

    def generic_visit(self, node):
        for field, value in ast.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        self.visit(item)
            elif isinstance(value, ast.AST):
                self.visit(value)

    def _report(self, node, message):
        self.findings.append(
            {
                "rule_id": "SKY-D231",
                "severity": "HIGH",
                "message": message,
                "file": str(self.file_path),
                "line": node.lineno,
                "col": node.col_offset,
            }
        )

    def visit_Assign(self, node):
        for target in node.targets:
            if isinstance(target, ast.Name):
                name = target.id
            elif isinstance(target, ast.Attribute):
                name = target.attr
            else:
                self.generic_visit(node)
                return

            if name == "CORS_ALLOW_ALL_ORIGINS" and isinstance(
                node.value, ast.Constant
            ):
                if node.value.value is True:
                    self._report(
                        node, "CORS misconfiguration: CORS_ALLOW_ALL_ORIGINS = True"
                    )

            if name in (
                "ACCESS_CONTROL_ALLOW_ORIGIN",
                "Access-Control-Allow-Origin",
            ):
                if isinstance(node.value, ast.Constant) and node.value.value == "*":
                    self._report(
                        node,
                        "CORS misconfiguration: Access-Control-Allow-Origin set to '*'",
                    )

        self.generic_visit(node)

    def visit_Call(self, node):
        qn = _qualified_name(node)
        if qn and qn.endswith("CORS"):
            has_origins = False
            for kw in node.keywords:
                if kw.arg in ("origins", "resources"):
                    has_origins = True
                    break
            if not has_origins and len(node.args) <= 1:
                self._report(
                    node,
                    "CORS misconfiguration: CORS() called without explicit origins restriction.",
                )

        self.generic_visit(node)

    def visit_Dict(self, node):
        for key, value in zip(node.keys, node.values):
            if (
                isinstance(key, ast.Constant)
                and key.value == "Access-Control-Allow-Origin"
                and isinstance(value, ast.Constant)
                and value.value == "*"
            ):
                self._report(
                    key,
                    "CORS misconfiguration: Access-Control-Allow-Origin set to '*'",
                )
        self.generic_visit(node)


def scan(tree, file_path, findings):
    try:
        checker = _CORSChecker(file_path, findings)
        checker.visit(tree)
    except Exception as e:
        print(f"CORS analysis failed for {file_path}: {e}", file=sys.stderr)



import json
from skylos.baseline import save_baseline, load_baseline, filter_new_findings


def _sample_result():
    return {
        "unused_functions": [
            {"name": "old_func", "file": "app.py", "line": 10},
            {"name": "another_func", "file": "app.py", "line": 20},
        ],
        "unused_imports": [
            {"name": "os", "file": "app.py", "line": 1},
        ],
        "unused_classes": [],
        "unused_variables": [],
        "danger": [
            {
                "rule_id": "SKY-D211",
                "file": "app.py",
                "line": 50,
                "message": "SQL injection",
            },
        ],
        "quality": [
            {
                "rule_id": "SKY-Q301",
                "file": "app.py",
                "line": 80,
                "message": "Complex function",
            },
        ],
        "secrets": [],
    }


class TestSaveBaseline:
    def test_creates_file(self, tmp_path):
        result = _sample_result()
        path = save_baseline(tmp_path, result)
        assert path.exists()
        assert path.name == "baseline.json"
        assert path.parent.name == ".skylos"

    def test_counts_correct(self, tmp_path):
        result = _sample_result()
        save_baseline(tmp_path, result)
        baseline = json.loads((tmp_path / ".skylos" / "baseline.json").read_text())
        assert baseline["counts"]["unused_functions"] == 2
        assert baseline["counts"]["unused_imports"] == 1
        assert baseline["counts"]["danger"] == 1
        assert baseline["counts"]["quality"] == 1
        assert baseline["counts"]["secrets"] == 0

    def test_fingerprints_created(self, tmp_path):
        result = _sample_result()
        save_baseline(tmp_path, result)
        baseline = json.loads((tmp_path / ".skylos" / "baseline.json").read_text())
        fps = set(baseline["fingerprints"])
        assert "dead:unused_functions:old_func" in fps
        assert "dead:unused_functions:another_func" in fps
        assert "dead:unused_imports:os" in fps
        assert "SKY-D211:app.py:50" in fps
        assert "SKY-Q301:app.py:80" in fps

    def test_overwrites_existing(self, tmp_path):
        save_baseline(tmp_path, _sample_result())
        save_baseline(
            tmp_path,
            {
                "unused_functions": [],
                "unused_imports": [],
                "unused_classes": [],
                "unused_variables": [],
                "danger": [],
                "quality": [],
                "secrets": [],
            },
        )
        baseline = json.loads((tmp_path / ".skylos" / "baseline.json").read_text())
        assert len(baseline["fingerprints"]) == 0


class TestLoadBaseline:
    def test_returns_none_if_missing(self, tmp_path):
        assert load_baseline(tmp_path) is None

    def test_loads_saved_baseline(self, tmp_path):
        save_baseline(tmp_path, _sample_result())
        baseline = load_baseline(tmp_path)
        assert baseline is not None
        assert "counts" in baseline
        assert "fingerprints" in baseline


class TestFilterNewFindings:
    def test_filters_known_findings(self):
        result = _sample_result()
        baseline = {
            "fingerprints": [
                "dead:unused_functions:old_func",
                "dead:unused_functions:another_func",
                "dead:unused_imports:os",
                "SKY-D211:app.py:50",
                "SKY-Q301:app.py:80",
            ]
        }
        filtered = filter_new_findings(result, baseline)
        assert len(filtered["unused_functions"]) == 0
        assert len(filtered["unused_imports"]) == 0
        assert len(filtered["danger"]) == 0
        assert len(filtered["quality"]) == 0

    def test_keeps_new_findings(self):
        result = _sample_result()
        result["danger"].append(
            {
                "rule_id": "SKY-D212",
                "file": "new.py",
                "line": 5,
                "message": "CMD injection",
            }
        )
        result["unused_functions"].append(
            {"name": "brand_new_func", "file": "new.py", "line": 15}
        )
        baseline = {
            "fingerprints": [
                "dead:unused_functions:old_func",
                "dead:unused_functions:another_func",
                "dead:unused_imports:os",
                "SKY-D211:app.py:50",
                "SKY-Q301:app.py:80",
            ]
        }
        filtered = filter_new_findings(result, baseline)
        assert len(filtered["danger"]) == 1
        assert filtered["danger"][0]["rule_id"] == "SKY-D212"
        assert len(filtered["unused_functions"]) == 1
        assert filtered["unused_functions"][0]["name"] == "brand_new_func"

    def test_empty_baseline_keeps_all(self):
        result = _sample_result()
        baseline = {"fingerprints": []}
        filtered = filter_new_findings(result, baseline)
        assert len(filtered["unused_functions"]) == 2
        assert len(filtered["danger"]) == 1
        assert len(filtered["quality"]) == 1

    def test_round_trip(self, tmp_path):
        """Save baseline, load it, filter — same result should yield 0 new."""
        result = _sample_result()
        save_baseline(tmp_path, result)
        baseline = load_baseline(tmp_path)
        filtered = filter_new_findings(result, baseline)
        total = sum(
            len(filtered.get(k, []))
            for k in [
                "unused_functions",
                "unused_imports",
                "unused_classes",
                "unused_variables",
                "danger",
                "quality",
                "secrets",
            ]
        )
        assert total == 0



from skylos.agent_center import rebuild_agent_state_from_existing, save_agent_state
from skylos.agent_service import AgentServiceController


def test_agent_service_controller_serves_command_center_and_triage_updates(tmp_path):
    project_root = tmp_path / "repo"
    project_root.mkdir()

    state = rebuild_agent_state_from_existing(
        {
            "project_root": str(project_root),
            "file_signatures": {"src/auth.py": {"mtime_ns": 1, "size": 10}},
            "changed_files": ["src/auth.py"],
            "baseline_present": False,
            "findings": [
                {
                    "fingerprint": "security:1",
                    "rule_id": "SKY-D999",
                    "category": "security",
                    "severity": "HIGH",
                    "message": "Shell injection path",
                    "file": "src/auth.py",
                    "absolute_file": str(project_root / "src" / "auth.py"),
                    "line": 9,
                    "confidence": 91,
                    "is_new_vs_baseline": True,
                    "is_new_since_last_scan": True,
                    "is_in_changed_file": True,
                },
            ],
        }
    )
    save_agent_state(project_root, state)

    controller = AgentServiceController(str(project_root))

    health = controller.health()
    assert health["ok"] is True
    assert health["has_state"] is True

    payload = controller.get_command_center()
    assert payload["items"][0]["id"] == "security:1"
    assert payload["triaged_count"] == 0

    dismissed = controller.dismiss("security:1")
    assert dismissed["triage"]["security:1"]["status"] == "dismissed"
    assert dismissed["actions"] == []

    restored = controller.restore("security:1")
    assert restored["triage"] == {}
    assert restored["actions"][0]["id"] == "security:1"


def test_agent_service_controller_tracks_snoozed_actions(tmp_path):
    project_root = tmp_path / "repo"
    project_root.mkdir()

    state = rebuild_agent_state_from_existing(
        {
            "project_root": str(project_root),
            "file_signatures": {"src/helpers.py": {"mtime_ns": 1, "size": 10}},
            "changed_files": [],
            "baseline_present": False,
            "findings": [
                {
                    "fingerprint": "dead:1",
                    "rule_id": "SKY-U002",
                    "category": "dead_code",
                    "severity": "INFO",
                    "message": "Unused import: os",
                    "file": "src/helpers.py",
                    "absolute_file": str(project_root / "src" / "helpers.py"),
                    "line": 2,
                    "confidence": 40,
                    "is_new_vs_baseline": False,
                    "is_new_since_last_scan": False,
                    "is_in_changed_file": False,
                },
            ],
        }
    )
    save_agent_state(project_root, state)

    controller = AgentServiceController(str(project_root))
    snoozed = controller.snooze("dead:1", hours=4)

    assert snoozed["triage"]["dead:1"]["status"] == "snoozed"
    assert snoozed["actions"] == []

    payload = controller.get_command_center()
    assert payload["triaged_count"] == 1
    assert payload["items"] == []



from __future__ import annotations

import ast
import pytest

import skylos.rules.danger.danger_web.xss_flow as xss


def _find_rule(findings, rule_id):
    for f in findings:
        if f.get("rule_id") == rule_id:
            return True
    return False


def test_qualified_name_from_call_simple():
    node = ast.parse("foo(1)").body[0].value
    assert isinstance(node, ast.Call)
    assert xss._qualified_name_from_call(node) == "foo"


def test_qualified_name_from_call_chain():
    node = ast.parse("a.b.c(1)").body[0].value
    assert isinstance(node, ast.Call)
    assert xss._qualified_name_from_call(node) == "a.b.c"


@pytest.mark.parametrize(
    "expr, expect",
    [
        ('f"hi {x}"', True),
        ('"a" + x', True),
        ('"<b>%s</b>" % x', True),
        ('"<b>{}</b>".format(x)', True),
        ('"plain"', False),
    ],
)
def test_is_interpolated_string(expr, expect):
    node = ast.parse(expr).body[0].value
    assert xss._is_interpolated_string(node) is expect


def test_const_str_value():
    assert xss._const_str_value(ast.parse('"hello"').body[0].value) == "hello"
    assert xss._const_str_value(ast.parse("123").body[0].value) is None


@pytest.mark.parametrize(
    "expr, expect",
    [
        ('"<b>hi</b>"', True),
        ('"nope"', False),
    ],
)
def test_const_contains_html(expr, expect):
    node = ast.parse(expr).body[0].value
    assert xss._const_contains_html(node) is expect


@pytest.mark.parametrize(
    "call_expr, expect",
    [
        ("Markup(user)", True),
        ('Markup(f"<b>{user}</b>")', True),
        ('Markup("<b>ok</b>")', False),
    ],
)
def test_rule_sky_d226(call_expr, expect):
    findings = []
    tree = ast.parse(f"def f(user):\n    return {call_expr}\n")
    xss.scan(tree, "x.py", findings)
    assert _find_rule(findings, "SKY-D226") is expect


def test_rule_sky_d226_mark_safe():
    findings = []
    tree = ast.parse("def f(user):\n    return mark_safe(user)\n")
    xss.scan(tree, "x.py", findings)
    assert _find_rule(findings, "SKY-D226") is True


@pytest.mark.parametrize(
    "tmpl, expect",
    [
        ('"hello |safe"', True),
        ('"{% autoescape false %}hi"', True),
        ('"{{ x }}"', False),
        ("tmpl", False),
    ],
)
def test_rule_sky_d227(tmpl, expect):
    findings = []
    tree = ast.parse(f"def f(tmpl):\n    return render_template_string({tmpl})\n")
    xss.scan(tree, "x.py", findings)
    assert _find_rule(findings, "SKY-D227") is expect


@pytest.mark.parametrize(
    "ret_expr, expect",
    [
        ('f"<b>{user}</b>"', True),
        ('f"hello {user}"', False),
        ('"<b>" + user + "</b>"', True),
        ('"<b>{}</b>".format(user)', True),
        ('"plain {}".format(user)', False),
        ('"<b>ok</b>" + clean', False),
    ],
)
def test_rule_sky_d228(ret_expr, expect):
    findings = []
    tree = ast.parse(f"def f(user):\n    return {ret_expr}\n")
    xss.scan(tree, "x.py", findings)
    assert _find_rule(findings, "SKY-D228") is expect


def test_rule_sky_d228_html_plus_local_clean_is_not_flagged():
    findings = []
    tree = ast.parse("def f(user):\n    clean = 'ok'\n    return '<b>ok</b>' + clean\n")
    xss.scan(tree, "x.py", findings)
    assert _find_rule(findings, "SKY-D228") is False


def test_scan_does_not_raise_on_exception(monkeypatch, capsys):
    class Boom(xss._XSSFlowChecker):
        def visit(self, node):
            raise RuntimeError("boom")

    monkeypatch.setattr(xss, "_XSSFlowChecker", Boom)

    findings = []
    tree = ast.parse("x = 1\n")
    xss.scan(tree, "x.py", findings)

    err = capsys.readouterr().err
    assert "XSS analysis failed for x.py" in err


def _scan_src(src: str):
    findings = []
    tree = ast.parse(src)
    xss.scan(tree, "x.py", findings)
    return findings


def _has_rule(findings, rule_id):
    for f in findings:
        if f.get("rule_id") == rule_id:
            return True
    return False


def test_rule_sky_d226_qualified_name_module_markup_flags():
    findings = _scan_src("def f(user):\n    return jinja2.Markup(user)\n")
    assert _has_rule(findings, "SKY-D226") is True


def test_rule_sky_d226_markup_with_local_const_is_not_flagged():
    findings = _scan_src("def f():\n    clean = 'ok'\n    return Markup(clean)\n")
    assert _has_rule(findings, "SKY-D226") is False


def test_rule_sky_d226_markup_interpolated_string_flags_even_if_local():
    findings = _scan_src(
        "def f():\n    clean = 'ok'\n    return Markup(f\"<b>{clean}</b>\")\n"
    )
    assert _has_rule(findings, "SKY-D226") is True


def test_rule_sky_d227_safe_filter_case_insensitive_flags():
    findings = _scan_src("def f():\n    return render_template_string('hello |SAFE')\n")
    assert _has_rule(findings, "SKY-D227") is True


def test_rule_sky_d227_autoescape_false_case_insensitive_flags():
    findings = _scan_src(
        "def f():\n    return render_template_string('{% AutoEscape False %}hi')\n"
    )
    assert _has_rule(findings, "SKY-D227") is True


def test_rule_sky_d228_fstring_html_with_local_not_tainted_not_flagged():
    findings = _scan_src("def f():\n    x = 'ok'\n    return f\"<b>{x}</b>\"\n")
    assert _has_rule(findings, "SKY-D228") is False


def test_rule_sky_d228_percent_format_html_with_taint_flags():
    findings = _scan_src("def f(user):\n    return '<b>%s</b>' % user\n")
    assert _has_rule(findings, "SKY-D228") is True


def test_rule_sky_d228_nested_binop_html_deeper_still_flags():
    findings = _scan_src("def f(user):\n    return '<b>' + ('x' + user)\n")
    assert _has_rule(findings, "SKY-D228") is True


def test_rule_sky_d228_nested_binop_html_on_right_still_flags():
    findings = _scan_src("def f(user):\n    return ('x' + user) + '</b>'\n")
    assert _has_rule(findings, "SKY-D228") is True


def test_rule_sky_d228_format_base_in_variable_not_detected_current_limitation():
    findings = _scan_src(
        "def f(user):\n    tmpl = '<b>{}</b>'\n    return tmpl.format(user)\n"
    )
    assert _has_rule(findings, "SKY-D228") is False



from __future__ import annotations

import json
import logging
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

logger = logging.getLogger(__name__)

FEEDBACK_DIR = Path.home() / ".skylos"
FEEDBACK_FILE = FEEDBACK_DIR / "feedback.json"

DEFAULT_WEIGHTS = {
    "same_file_attr": 1.0,
    "same_pkg_attr": 0.3,
    "global_attr": 0.1,
}

MIN_WEIGHT = 0.02
MAX_WEIGHT = 2.0
MIN_OBSERVATIONS = 5


@dataclass
class HeuristicObservation:
    heuristic_type: str
    is_spurious: bool
    function_name: str = ""
    project: str = ""


@dataclass
class FeedbackData:
    observations: dict[str, dict[str, int]] = field(default_factory=dict)
    tuned_weights: dict[str, float] = field(default_factory=dict)
    total_runs: int = 0

    def to_dict(self) -> dict:
        return {
            "observations": self.observations,
            "tuned_weights": self.tuned_weights,
            "total_runs": self.total_runs,
        }

    @classmethod
    def from_dict(cls, data: dict) -> FeedbackData:
        return cls(
            observations=data.get("observations", {}),
            tuned_weights=data.get("tuned_weights", {}),
            total_runs=data.get("total_runs", 0),
        )


def load_feedback() -> FeedbackData:
    if not FEEDBACK_FILE.exists():
        return FeedbackData()
    try:
        data = json.loads(FEEDBACK_FILE.read_text(encoding="utf-8"))
        return FeedbackData.from_dict(data)
    except (json.JSONDecodeError, Exception) as e:
        logger.warning(f"Failed to load feedback: {e}")
        return FeedbackData()


def save_feedback(data: FeedbackData) -> None:
    try:
        FEEDBACK_DIR.mkdir(parents=True, exist_ok=True)
        FEEDBACK_FILE.write_text(
            json.dumps(data.to_dict(), indent=2),
            encoding="utf-8",
        )
    except Exception as e:
        logger.warning(f"Failed to save feedback: {e}")


def record_verification_results(
    verification_result: dict[str, Any],
) -> FeedbackData:
    feedback = load_feedback()

    for finding in verification_result.get("verified_findings", []):
        verdict = finding.get("_llm_verdict", "")
        heuristic_refs = finding.get("heuristic_refs", {})

        if not heuristic_refs:
            continue
        if verdict not in ("TRUE_POSITIVE", "FALSE_POSITIVE"):
            continue

        for htype in heuristic_refs:
            if htype not in feedback.observations:
                feedback.observations[htype] = {
                    "real": 0,
                    "spurious": 0,
                    "uncertain": 0,
                }

            if verdict == "TRUE_POSITIVE":
                feedback.observations[htype]["spurious"] += 1
            elif verdict == "FALSE_POSITIVE":
                feedback.observations[htype]["real"] += 1

    for new_dead in verification_result.get("new_dead_code", []):
        heuristic_refs = new_dead.get("heuristic_refs", {})
        for htype in heuristic_refs:
            if htype not in feedback.observations:
                feedback.observations[htype] = {
                    "real": 0,
                    "spurious": 0,
                    "uncertain": 0,
                }
            feedback.observations[htype]["spurious"] += 1

    feedback.total_runs += 1

    feedback.tuned_weights = compute_tuned_weights(feedback)

    save_feedback(feedback)
    return feedback


def compute_tuned_weights(feedback: FeedbackData) -> dict[str, float]:
    tuned = {}

    for htype, default_w in DEFAULT_WEIGHTS.items():
        obs = feedback.observations.get(htype, {})
        real = obs.get("real", 0)
        spurious = obs.get("spurious", 0)
        total = real + spurious

        if total < MIN_OBSERVATIONS:
            tuned[htype] = default_w
            continue

        accuracy = real / total
        new_weight = default_w * accuracy
        new_weight = max(MIN_WEIGHT, min(MAX_WEIGHT, new_weight))

        tuned[htype] = round(new_weight, 4)

    return tuned


def get_tuned_weights() -> dict[str, float]:
    feedback = load_feedback()
    weights = dict(DEFAULT_WEIGHTS)
    weights.update(feedback.tuned_weights)
    return weights


def get_feedback_summary() -> dict[str, Any]:
    feedback = load_feedback()

    summary = {
        "total_runs": feedback.total_runs,
        "heuristic_types": {},
    }

    for htype in DEFAULT_WEIGHTS:
        obs = feedback.observations.get(htype, {})
        real = obs.get("real", 0)
        spurious = obs.get("spurious", 0)
        total = real + spurious

        default_w = DEFAULT_WEIGHTS[htype]
        tuned_w = feedback.tuned_weights.get(htype, default_w)

        summary["heuristic_types"][htype] = {
            "observations": total,
            "real": real,
            "spurious": spurious,
            "accuracy_pct": round(100 * real / total, 1) if total > 0 else None,
            "default_weight": default_w,
            "tuned_weight": tuned_w,
            "weight_change_pct": round(100 * (tuned_w - default_w) / default_w, 1)
            if default_w > 0
            else 0,
        }

    return summary


def reset_feedback() -> None:
    if FEEDBACK_FILE.exists():
        FEEDBACK_FILE.unlink()



import json
from pathlib import Path
import pytest
import skylos.sync as syncmod
import builtins


class FakeResponse:
    def __init__(self, status_code=200, payload=None, text="OK", raise_exc=None):
        self.status_code = status_code
        self._payload = payload if payload is not None else {}
        self.text = text
        self._raise_exc = raise_exc

    def json(self):
        return self._payload

    def raise_for_status(self):
        if self._raise_exc:
            raise self._raise_exc


@pytest.fixture()
def isolated_creds(monkeypatch, tmp_path):
    monkeypatch.delenv("SKYLOS_TOKEN", raising=False)
    monkeypatch.delenv("SKYLOS_API_URL", raising=False)

    home_dir = tmp_path / "home"
    home_dir.mkdir(parents=True, exist_ok=True)

    creds_dir = home_dir / ".skylos"
    creds_file = creds_dir / "credentials.json"

    monkeypatch.setattr(syncmod, "GLOBAL_CREDS_DIR", creds_dir, raising=False)
    monkeypatch.setattr(syncmod, "GLOBAL_CREDS_FILE", creds_file, raising=False)

    return creds_dir, creds_file


def test_mask_token_short():
    assert syncmod.mask_token("abc") == "****"
    assert syncmod.mask_token("") == "****"
    assert syncmod.mask_token(None) == "****"


def test_mask_token_long():
    tok = "skylos_token_1234567890ABCDEFG"
    masked = syncmod.mask_token(tok)
    assert masked.startswith(tok[:8] + "...")
    assert masked.endswith(tok[-4:])
    assert tok not in masked


def test_get_token_env_wins(isolated_creds, monkeypatch):
    _, creds_file = isolated_creds
    creds_file.parent.mkdir(parents=True, exist_ok=True)
    creds_file.write_text(json.dumps({"token": "FILE_TOKEN"}))

    monkeypatch.setenv("SKYLOS_TOKEN", "ENV_TOKEN")
    assert syncmod.get_token() == "ENV_TOKEN"


def test_get_token_from_global_creds_file(isolated_creds):
    _, creds_file = isolated_creds
    creds_file.parent.mkdir(parents=True, exist_ok=True)
    creds_file.write_text(json.dumps({"token": "FILE_TOKEN"}))
    assert syncmod.get_token() == "FILE_TOKEN"


def test_get_token_none_if_missing(isolated_creds):
    _, creds_file = isolated_creds
    assert not creds_file.exists()
    assert syncmod.get_token() is None


def test_save_token_writes_file(isolated_creds):
    _, creds_file = isolated_creds
    assert not creds_file.exists()

    out_path = syncmod.save_token(
        "TOK_123",
        project_id="proj_abc",
        project_name="Proj",
        org_name="Org",
        plan="pro",
    )

    assert out_path == str(creds_file)
    assert creds_file.exists()
    data = json.loads(creds_file.read_text())
    assert data["token"] == "TOK_123"
    assert data["plan"] == "pro"
    assert data["saved_at"].endswith("Z")
    assert data["tokens"]["proj_abc"]["project_name"] == "Proj"
    assert data["tokens"]["proj_abc"]["org_name"] == "Org"


def test_clear_token(isolated_creds):
    _, creds_file = isolated_creds
    creds_file.parent.mkdir(parents=True, exist_ok=True)
    creds_file.write_text("{}")

    assert syncmod.clear_token() is True
    assert not creds_file.exists()
    assert syncmod.clear_token() is False


def test_api_get_success(monkeypatch):
    def fake_get(url, headers=None, timeout=None):
        assert "Authorization" in (headers or {})
        return FakeResponse(200, {"ok": True})

    monkeypatch.setattr(syncmod.requests, "get", fake_get)
    monkeypatch.setenv("SKYLOS_API_URL", "https://example.com")
    out = syncmod.api_get("/api/sync/whoami", "TOKEN")
    assert out == {"ok": True}


def test_api_get_401_raises(monkeypatch):
    def fake_get(url, headers=None, timeout=None):
        return FakeResponse(401, {"ok": False})

    monkeypatch.setattr(syncmod.requests, "get", fake_get)
    with pytest.raises(syncmod.AuthError) as e:
        syncmod.api_get("/api/sync/whoami", "BADTOKEN")
    assert "Invalid API token" in str(e.value)


def test_api_get_connection_error(monkeypatch):
    def fake_get(url, headers=None, timeout=None):
        raise syncmod.requests.exceptions.ConnectionError()

    monkeypatch.setattr(syncmod.requests, "get", fake_get)
    with pytest.raises(syncmod.AuthError) as e:
        syncmod.api_get("/api/sync/whoami", "TOKEN")
    assert "Cannot connect" in str(e.value)


def test_api_get_timeout(monkeypatch):
    def fake_get(url, headers=None, timeout=None):
        raise syncmod.requests.exceptions.Timeout()

    monkeypatch.setattr(syncmod.requests, "get", fake_get)
    with pytest.raises(syncmod.AuthError) as e:
        syncmod.api_get("/api/sync/whoami", "TOKEN")
    assert "Request timed out" in str(e.value)


def test_cmd_status_not_connected(isolated_creds, capsys):
    syncmod.cmd_status()
    out = capsys.readouterr().out
    assert "Not connected" in out
    assert "skylos sync connect" in out


def test_cmd_status_connected_ok(isolated_creds, monkeypatch, capsys):
    _, creds_file = isolated_creds
    creds_file.parent.mkdir(parents=True, exist_ok=True)
    creds_file.write_text(json.dumps({"token": "TOK"}))

    def fake_api_get(endpoint, token):
        assert endpoint == "/api/sync/whoami"
        assert token == "TOK"
        return {
            "project": {"name": "MyProj"},
            "organization": {"name": "MyOrg"},
            "plan": "free",
        }

    monkeypatch.setattr(syncmod, "api_get", fake_api_get)

    syncmod.cmd_status()
    out = capsys.readouterr().out
    assert "✓ Connected" in out
    assert "Project:" in out and "MyProj" in out
    assert "Organization:" in out and "MyOrg" in out
    assert "Plan:" in out and "Free" in out


def test_cmd_disconnect(isolated_creds, capsys):
    _, creds_file = isolated_creds
    creds_file.parent.mkdir(parents=True, exist_ok=True)
    creds_file.write_text("{}")

    syncmod.cmd_disconnect()
    out = capsys.readouterr().out
    assert "Disconnected" in out

    syncmod.cmd_disconnect()
    out2 = capsys.readouterr().out
    assert "No saved credentials" in out2


def test_cmd_connect_with_token_arg_saves_creds(isolated_creds, monkeypatch, capsys):
    def fake_api_get(endpoint, token):
        assert endpoint == "/api/sync/whoami"
        assert token == "TOK_ARG"
        return {
            "project": {"id": "proj_123", "name": "Proj"},
            "organization": {"name": "Org"},
            "plan": "pro",
        }

    monkeypatch.setattr(syncmod, "api_get", fake_api_get)

    syncmod.cmd_connect("TOK_ARG")
    out = capsys.readouterr().out

    assert "Verifying token" in out
    assert "✓ Connected!" in out
    assert "Project:" in out and "Proj" in out
    assert "Organization:" in out and "Org" in out
    assert "Plan:" in out and "Pro" in out

    _, creds_file = isolated_creds
    assert creds_file.exists()
    data = json.loads(creds_file.read_text())
    assert data["token"] == "TOK_ARG"
    assert data["plan"] == "pro"
    assert data["tokens"]["proj_123"]["project_name"] == "Proj"
    assert data["tokens"]["proj_123"]["org_name"] == "Org"


def test_cmd_connect_cancel_input(monkeypatch):
    monkeypatch.delenv("SKYLOS_TOKEN", raising=False)

    def _raise_keyboard_interrupt(_prompt=""):
        raise KeyboardInterrupt

    monkeypatch.setattr(builtins, "input", _raise_keyboard_interrupt)

    with pytest.raises(SystemExit) as e:
        syncmod.cmd_connect(None)

    assert e.value.code == 1


def test_cmd_pull_not_connected_exits(isolated_creds, capsys):
    with pytest.raises(SystemExit) as e:
        syncmod.cmd_pull()
    assert e.value.code == 1
    out = capsys.readouterr().out
    assert "Not connected" in out or "Run 'skylos sync connect'" in out


def test_cmd_pull_writes_config_and_suppressions(
    isolated_creds, monkeypatch, tmp_path, capsys
):
    _, creds_file = isolated_creds
    creds_file.parent.mkdir(parents=True, exist_ok=True)
    creds_file.write_text(json.dumps({"token": "TOK"}))

    monkeypatch.setattr(syncmod, "SKYLOS_DIR", str(tmp_path / ".skylos"), raising=False)

    def fake_api_get(endpoint, token):
        assert token == "TOK"
        if endpoint == "/api/sync/whoami":
            return {"project": {"name": "Proj"}}
        if endpoint == "/api/sync/config":
            return {"config": {"complexity": 12, "nesting": 4}}
        if endpoint == "/api/sync/suppressions":
            return {"suppressions": [{"rule_id": "SKY-D212"}], "count": 1}
        raise AssertionError(f"Unexpected endpoint {endpoint}")

    monkeypatch.setattr(syncmod, "api_get", fake_api_get)

    syncmod.cmd_pull()
    out = capsys.readouterr().out

    assert "Pulling configuration" in out
    assert "Pulling suppressions" in out
    assert "Sync complete" in out

    skylos_dir = Path(syncmod.SKYLOS_DIR)
    config_path = skylos_dir / syncmod.CONFIG_FILE
    supp_path = skylos_dir / syncmod.SUPPRESSIONS_FILE

    assert config_path.exists()
    assert supp_path.exists()

    config_text = config_path.read_text()
    assert "complexity" in config_text
    assert "nesting" in config_text

    supp = json.loads(supp_path.read_text())
    assert isinstance(supp, list)
    assert supp[0]["rule_id"] == "SKY-D212"


def test_main_usage_no_args(capsys):
    syncmod.main([])
    out = capsys.readouterr().out
    assert "Usage: skylos sync <command>" in out
    assert "connect" in out
    assert "pull" in out


def test_main_unknown_command_exits(capsys):
    with pytest.raises(SystemExit) as e:
        syncmod.main(["wat"])
    assert e.value.code == 1
    out = capsys.readouterr().out
    assert "Unknown command" in out


def test_main_dispatch_connect(monkeypatch):
    called = {"ok": False}

    def fake_connect(arg):
        called["ok"] = True
        assert arg == "T"

    monkeypatch.setattr(syncmod, "cmd_connect", fake_connect)
    syncmod.main(["connect", "T"])
    assert called["ok"] is True
